/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Veerapuram Varadhan <v.varadhan@gmail.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "ldtp.h"
#include "remap.h"
#include "ldtp-gui.h"
#include "ldtp-state.h"
#include "ldtp-utils.h"
#include "ldtp-logger.h"
#include "ldtp-memory.h"
#include "ldtp-record.h"
#include "ldtp-command.h"
#include "localization.h"
#include "ldtp-gui-comp.h"
#include "client-handler.h"

extern GHashTable *event_notifier;

static void
register_window_creation_event (LDTPClientContext* cctxt, LDTPError** err)
{
	if (!cctxt || !cctxt->req || !cctxt->req->context)
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);

	if (!event_notifier)
		event_notifier = g_hash_table_new (&g_str_hash, &g_str_equal);

	g_hash_table_insert (event_notifier, g_strdup ((gchar *)cctxt->req->context), cctxt);
	*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
}

void
generate_response_packet (LDTPClientContext* cctxt, LDTPError** err, char **resp_pckt, uint32_t *resp_size)
{
	uint32_t resp_len = 0;
	char buf [16];  
	char buf1 [16];
	char *resp_xml = NULL;
	char *data_pckt = NULL;

	if (!cctxt || !cctxt->resp) {
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		return;
	}

	resp_len = XML_HEADER_LEN + 
		RESPONSE_ELEMENT_LEN + 
		RESPONSE_ID_ELEMENT_LEN +
		RESPONSE_ID_END_ELEMENT_LEN +
		STATUS_ELEMENT_LEN +
		ATTRIBUTE_CODE_ELEMENT_LEN +
		ATTRIBUTE_CODE_END_ELEMENT_LEN +
		ATTRIBUTE_MSG_ELEMENT_LEN +
		ATTRIBUTE_MSG_END_ELEMENT_LEN +
		STATUS_END_ELEMENT_LEN +
		RESPONSE_END_ELEMENT_LEN;

	/* Varadhan
	   FIXME: Localization?? 
	*/

	g_print ("resp_len = %d\n", resp_len);

	if (cctxt->req->request_id)
		resp_len += g_utf8_strlen ((gchar *)cctxt->req->request_id, -1);
	if (cctxt->resp->resp_status && cctxt->resp->resp_status->message)
		resp_len += g_utf8_strlen (cctxt->resp->resp_status->message, -1);

	memset (buf, 0x00, 16);
	sprintf (buf, "%d", cctxt->resp->resp_status->err_code);
	resp_len += strlen (buf);

	if (cctxt->resp->data && cctxt->resp->data_len > 0) {
		resp_len += DATA_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_END_ELEMENT_LEN +
			ATTRIBUTE_VALUE_ELEMENT_LEN +
			ATTRIBUTE_VALUE_END_ELEMENT_LEN +
			DATA_END_ELEMENT_LEN +
			cctxt->resp->data_len;
		memset (buf1, 0x00, 16);
		sprintf (buf1, "%d", cctxt->resp->data_len);
		resp_len += strlen (buf1);
	}

	if (cctxt->resp->data && cctxt->resp->data_len > 0) 
		data_pckt = g_strconcat (DATA_ELEMENT, 
					 ATTRIBUTE_LENGTH_ELEMENT,
					 buf1, ATTRIBUTE_LENGTH_END_ELEMENT,
					 ATTRIBUTE_VALUE_ELEMENT, 
					 cctxt->resp->data,
					 ATTRIBUTE_VALUE_END_ELEMENT, 
					 DATA_END_ELEMENT, NULL);

	resp_xml = g_strdup_printf ("%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
				    XML_HEADER, RESPONSE_ELEMENT,
				    RESPONSE_ID_ELEMENT,
				    cctxt->req->request_id,
				    RESPONSE_ID_END_ELEMENT,
				    STATUS_ELEMENT, ATTRIBUTE_CODE_ELEMENT,
				    buf, ATTRIBUTE_CODE_END_ELEMENT,
				    ATTRIBUTE_MSG_ELEMENT, 
				    cctxt->resp->resp_status->message,
				    ATTRIBUTE_MSG_END_ELEMENT,
				    STATUS_END_ELEMENT,
				    data_pckt ? data_pckt : "",
				    RESPONSE_END_ELEMENT);

	*resp_size = strlen (resp_xml);
	*resp_pckt = resp_xml;
}

void
generate_record_response_packet (LDTPClientContext* cctxt,
				 LDTPError** err,
				 char **resp_pckt,
				 uint32_t *resp_size)
{
	uint32_t resp_len = 0;
	char buf [16];  
	char buf1 [16];  
	char *resp_xml = NULL;
	char *data_pckt = NULL;

	if (!cctxt || !cctxt->resp) {
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		return;
	}

	resp_len = XML_HEADER_LEN + 
		RECORD_ELEMENT_LEN + 
		RECORD_END_ELEMENT_LEN;
	/* Varadhan
	   FIXME: Localization?? 
	*/
	if (cctxt->req->request_id)
		resp_len += g_utf8_strlen ((gchar *)cctxt->req->request_id, -1);
	if (cctxt->resp->resp_status && cctxt->resp->resp_status->message)
		resp_len += g_utf8_strlen (cctxt->resp->resp_status->message, -1);

	memset (buf, 0x00, 16);
	sprintf (buf, "%d", cctxt->resp->resp_status->err_code);
	resp_len += strlen (buf);

	if (cctxt->resp->data && cctxt->resp->data_len > 0) {
		resp_len += RESPONSE_ID_ELEMENT_LEN + 
			RESPONSE_ID_END_ELEMENT_LEN +
			DATA_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_END_ELEMENT_LEN +
			RECORD_FILE_ELEMENT_LEN +
			RECORD_FILE_END_ELEMENT_LEN +
			DATA_END_ELEMENT_LEN +
			cctxt->resp->data_len;
		memset (buf1, 0x00, 16);
		sprintf (buf1, "%d", cctxt->resp->data_len);
		resp_len += strlen (buf1);
	}

	if (cctxt->resp->data && cctxt->resp->data_len > 0) 
		data_pckt = g_strconcat (DATA_ELEMENT,
					 ATTRIBUTE_LENGTH_ELEMENT,
					 buf1, ATTRIBUTE_LENGTH_END_ELEMENT,
					 RECORD_FILE_ELEMENT, 
					 cctxt->resp->data,
					 RECORD_FILE_END_ELEMENT, 
					 DATA_END_ELEMENT, NULL);

	resp_xml = g_strdup_printf ("%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
				    XML_HEADER,
				    RECORD_ELEMENT,
				    RESPONSE_ID_ELEMENT,
				    cctxt->req->request_id,
				    RESPONSE_ID_END_ELEMENT,
				    STATUS_ELEMENT, ATTRIBUTE_CODE_ELEMENT,
				    buf, ATTRIBUTE_CODE_END_ELEMENT,
				    ATTRIBUTE_MSG_ELEMENT, 
				    cctxt->resp->resp_status->message,
				    ATTRIBUTE_MSG_END_ELEMENT,
				    STATUS_END_ELEMENT,
				    data_pckt != NULL ? data_pckt : "",
				    RECORD_END_ELEMENT);
	*resp_size = strlen (resp_xml);
	*resp_pckt = resp_xml;
}

void
generate_notification_packet (LDTPClientContext* cctxt, LDTPError** err, char **resp_pckt, uint32_t *resp_size)
{
	uint32_t resp_len = 0;
	char buf [16];  
	char buf1 [16];
	char *resp_xml = NULL;
	char *data_pckt = NULL;

	if (!cctxt || !cctxt->resp) {
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		return;
	}

	resp_len = XML_HEADER_LEN + 
		NOTIFICATION_ELEMENT_LEN + 
		STATUS_ELEMENT_LEN +
		ATTRIBUTE_CODE_ELEMENT_LEN +
		ATTRIBUTE_CODE_END_ELEMENT_LEN +
		ATTRIBUTE_MSG_ELEMENT_LEN +
		ATTRIBUTE_MSG_END_ELEMENT_LEN +
		STATUS_END_ELEMENT_LEN +
		NOTIFICATION_END_ELEMENT_LEN;

	/* Varadhan
	   FIXME: Localization?? 
	*/

	g_print ("resp_len = %d\n", resp_len);

	resp_len += strlen (cctxt->resp->resp_status->message);

	memset (buf, 0x00, 16);  
	sprintf (buf, "%d", cctxt->resp->resp_status->err_code);
	resp_len += strlen (buf);

	if (cctxt->resp->data && cctxt->resp->data_len > 0) {
		resp_len += DATA_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_ELEMENT_LEN +
			ATTRIBUTE_LENGTH_END_ELEMENT_LEN +
			ATTRIBUTE_VALUE_ELEMENT_LEN +
			ATTRIBUTE_VALUE_END_ELEMENT_LEN +
			DATA_END_ELEMENT_LEN +
			cctxt->resp->data_len;
		memset (buf1, 0x00, 16);
		sprintf (buf1, "%d", cctxt->resp->data_len);
		resp_len += strlen (buf1);
	}

	if (cctxt->resp->data && cctxt->resp->data_len > 0) 
		data_pckt = g_strconcat (DATA_ELEMENT, 
					 ATTRIBUTE_LENGTH_ELEMENT,
					 buf1, ATTRIBUTE_LENGTH_END_ELEMENT,
					 ATTRIBUTE_VALUE_ELEMENT, 
					 cctxt->resp->data,
					 ATTRIBUTE_VALUE_END_ELEMENT, 
					 DATA_END_ELEMENT, NULL);

	resp_xml = g_strdup_printf ("%s%s%s%s%s%s%s%s%s%s%s%s",
				    XML_HEADER, NOTIFICATION_ELEMENT,
				    STATUS_ELEMENT, ATTRIBUTE_CODE_ELEMENT,
				    buf, ATTRIBUTE_CODE_END_ELEMENT,
				    ATTRIBUTE_MSG_ELEMENT, 
				    cctxt->resp->resp_status->message,
				    ATTRIBUTE_MSG_END_ELEMENT,
				    STATUS_END_ELEMENT,
				    data_pckt ? data_pckt : "",
				    NOTIFICATION_END_ELEMENT);

	*resp_size = strlen (resp_xml);
	*resp_pckt = resp_xml;
}

void
send_response (int sockfd, char *resp_xml, uint32_t resp_len, LDTPError **err)
{
	char* resp_packet = NULL;
	//int i = 0;
	int pckt_len = 0;
	uint32_t resp_size = 0;

	resp_size = htonl (resp_len);
	resp_len += sizeof (resp_len);
	resp_packet = calloc (resp_len + 1, 1);
	memcpy (resp_packet, (void *)&resp_size, sizeof (resp_size));

	uint32_t temp = ntohl (resp_size);
	g_print ("Sending..\n%d\n", temp);

	memcpy ((resp_packet + sizeof (resp_len)), 
		resp_xml, resp_len - sizeof (resp_len));

	g_print ("Response packet: %s\n", (resp_packet + sizeof (resp_len)));

	pckt_len = resp_len;

	/*
	  If data sent in chunks from server to client, then the peek code
	  in client was not able to continue reading the next chunk as the
	  recv with peek option always returns the first chunk.
	*/

	ldtp_send (sockfd, resp_packet, resp_len, err);

	/*
	while (pckt_len > 0) {
		pckt_len = pckt_len > 512 ? 512 : pckt_len;
		ldtp_send (sockfd, resp_packet+i, pckt_len, err);
		if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
			// FIXME: Error handling
			g_print ("Error: %s\n", (*err)->message);
			break;
		}
		i += pckt_len;
		pckt_len = resp_len - i;
	}
	*/
	
	/*
	  FIXME: Do not fragment memory.  Use something similar to
	  res-packet thingy.
	*/
	g_free (resp_packet);
}

static void
add_item_to_list (gpointer key, gpointer value, gpointer list)
{
        char **user_data = list;
        if (!key)
                return;
	char *data = *user_data;
        if (data) {
                gchar *tmp = NULL;
                gchar *escaped_xml_data = NULL;
                escaped_xml_data = escape_xml_character (key);

                /*
                  FIXME: Avoid memory fragmentation
		*/
                tmp = g_strdup_printf ("%s<OBJECT>%s</OBJECT>", data, escaped_xml_data);
                g_free (escaped_xml_data);
                g_free (data);
                data = tmp;
        }
        else
                data = g_strdup_printf ("<OBJECT>%s</OBJECT>", (char *) key);
	*user_data = data;
}

#ifdef ENABLE_LOCALIZATION
static void
createpo (gchar *package_name, gchar *locale_dir, gchar *file_name,
	  gchar *mode, FILE *log_fp, LDTPError **err)
{
	gchar *locale_lang = NULL;
	locale_lang = getenv ("LANG");
	if (locale_lang == NULL) {
		log_msg (LDTP_LOG_CAUSE, "Locale language environment not set - LANG",
			 log_fp);
		*err = ldtp_error_get_error (LDTP_ERROR_UTF8_ENGLISH_LANG);
		return;
	}
	else {
		gchar *tmp = NULL;
		gchar *mofilename = NULL;
		gchar *mopathfilename = NULL;

		struct stat dir_buf;
		gboolean folder_found = FALSE;

		if (g_ascii_strcasecmp (locale_lang, "en_US.UTF-8") == 0) {
			*err = ldtp_error_get_error (LDTP_ERROR_UTF8_ENGLISH_LANG);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		if (stat (locale_dir, &dir_buf) != 0) {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_STAT_DIR);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		if (!S_ISDIR (dir_buf.st_mode)) {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_STAT_DIR);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		if (g_ascii_strcasecmp (mode, "mo") != 0) {
			*err = ldtp_error_get_error (LDTP_ERROR_ONLY_MO_MODE_SUPPORTED);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		tmp = g_strdup_printf ("%s/%s", locale_dir, locale_lang);
		if (!tmp) {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		mofilename = g_strdup_printf ("/LC_MESSAGES/%s.mo", package_name);
		if (!mofilename) {
			g_free (tmp);
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
			return;
		}
		g_print ("Checking in %s%s\n", tmp, mofilename);

		if ((stat (tmp, &dir_buf) == 0) && S_ISDIR (dir_buf.st_mode)) {
			mopathfilename = g_strdup_printf ("%s%s", tmp, mofilename);
			if (!mopathfilename) {
				g_free (tmp);
				g_free (mofilename);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("MO path filename: %s\n", mopathfilename);
			if ((stat (mopathfilename, &dir_buf) == 0) && S_ISREG (dir_buf.st_mode)) {
				folder_found = TRUE;
			}
		}
		if (!folder_found && g_utf8_strchr (tmp, -1, '.')) {
			gchar *stripped_data = strip_delim (tmp, '.');
			g_free (mopathfilename);
			if (!stripped_data) {
				g_free (tmp);
				g_free (mofilename);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("MO Path sripped: %s\n", stripped_data);
			mopathfilename = g_strdup_printf ("%s%s", stripped_data, mofilename);
			g_free (stripped_data);
			if (!mopathfilename) {
				g_free (tmp);
				g_free (mofilename);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("MO path filename: %s\n", mopathfilename);
			if ((stat (mopathfilename, &dir_buf) == 0) && S_ISREG (dir_buf.st_mode)) {
				folder_found = TRUE;
				log_msg (LDTP_LOG_INFO, "Folder found", log_fp);
			}
		}
		if (!folder_found && g_utf8_strchr (tmp, -1, '_')) {
			gchar *stripped_data = strip_delim (tmp, '_');
			g_free (mopathfilename);
			if (!stripped_data) {
				g_free (tmp);
				g_free (mofilename);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("MO Path sripped: %s\n", stripped_data);
			mopathfilename = g_strdup_printf ("%s%s", stripped_data, mofilename);
			g_free (stripped_data);
			if (!mopathfilename) {
				g_free (tmp);
				g_free (mofilename);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("MO path filename: %s\n", mopathfilename);
			if ((stat (mopathfilename, &dir_buf) == 0) && S_ISREG (dir_buf.st_mode)) {
				folder_found = TRUE;
			}
		}
		g_free (tmp);
		tmp = NULL;
		g_free (mofilename);
		mofilename = NULL;
		if (folder_found) {
			FILE *fp;
			gchar *unformat = NULL;
			gchar *data = NULL;

			log_msg (LDTP_LOG_INFO, mopathfilename, log_fp);
			unformat = g_strdup_printf ("msgunfmt %s -o %s", mopathfilename, file_name);
			g_free (mopathfilename);
			mopathfilename = NULL;
			if (!unformat) {
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			g_print ("Command: %s\n", unformat);
			fp = popen (unformat, "r");
			if (!fp) {
				g_free (unformat);
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_ALLOCATE_MEMORY);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
				return;
			}
			data = read_line (fileno (fp), err);
			if (data)
				g_print ("MO unformat command output: %s\n", data);
			g_free (data);
			g_free (unformat);
			pclose (fp);
			log_msg (LDTP_LOG_INFO, "Success", log_fp);
			return;
		}
		g_free (mopathfilename);
	}
}
#endif

#ifdef ENABLE_LOCALIZATION
static void
deletepo (char *filename, FILE *log_fp, LDTPError **err)
{
	if (unlink (filename) == 0) {
		log_msg (LDTP_LOG_INFO, "Successfully deleted file", log_fp);
		*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
		return;
	}
	else {
		*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_DELETE_PO);
		log_msg (LDTP_LOG_CAUSE, (*err)->message, log_fp);
		return;
	}
}
#endif

static void
bind_text (LDTPClientContext *cctxt, gchar *mode, LDTPError** err)
{
#ifdef ENABLE_LOCALIZATION
	gboolean temp_flag;
	gchar *filename = NULL;
	gchar *package_name = cctxt->req->context;
	gchar *locale_dir = cctxt->req->component;

	bindtextdomain (package_name, locale_dir);
	bind_textdomain_codeset (package_name, "UTF-8");
	textdomain (package_name);
	/*
	 * FIXME
	 * Currently only mo files are handled
	 */
	filename = g_strdup_printf ("%s/%s-%d.po", g_get_tmp_dir (),
				    package_name, g_random_int_range (G_MININT, G_MAXINT));
	createpo (package_name, locale_dir, filename, mode, cctxt->log_fp, err);
	if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
		*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_CREATE_PO);
		log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
		return;
	}
	temp_flag = init_catalog (filename, cctxt->log_fp);
	if (!cctxt->locale_set && temp_flag)
		cctxt->locale_set = temp_flag;
	deletepo (filename, cctxt->log_fp, err);
	g_free (filename);
#else
	g_print ("Localization option not enabled in LDTP\n");
	*err = ldtp_error_get_error (LDTP_ERROR_INVALID_COMMAND);
#endif
}

static void
has_state (LDTPClientContext* cctxt,
	   LDTPError** err)
{
	guint i;
	char *state = NULL;
	int state_id;
	AccessibleStateSet *state_set;
	guint len = g_slist_length (cctxt->req->arg_list);

	state_set = Accessible_getStateSet (cctxt->gui_handle->handle);

	if (len == 0) {
		*err = ldtp_error_get_error (LDTP_ERROR_INVALID_STATE);
		goto error;
	}

	for (i = 0; i < len; i++) {
		state = g_slist_nth_data (cctxt->req->arg_list, i);
		g_print ("Command: %s\n", state);
		state_id = ldtp_state_get_state_id ((gchar *)state);
		if (state_id == LDTP_STATE_INVALID) {
			*err = ldtp_error_get_error (LDTP_ERROR_INVALID_STATE);
			goto error;
		}
		if (AccessibleStateSet_contains (state_set, state_id) == FALSE) {
			*err = ldtp_error_get_error (LDTP_ERROR_INVALID_STATE);
			goto error;
		}
	}
	*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
 error:
	ldtp_gui_free_gui_handle (cctxt->gui_handle);
	cctxt->gui_handle = NULL;
}

static void
handle_request (LDTPClientContext* cctxt, 
		Packet* pckt, LDTPError** err)
{
	int command;
	static FILE *tmp_fp = NULL;

	LDTPRequest* ldtp_req = NULL;
	//LDTPResponse* ldtp_resp = NULL;
	LDTPGuiHandle* accessible = NULL;

	ldtp_req = cctxt->req;
	ldtp_request_fill_request (ldtp_req, pckt->packet, pckt->len, err);
	if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
		/* FIXME: Error handling */
		return;
	}

	g_print ("Command: %s\n", ldtp_req->action_name);
	command = ldtp_command_get_command_id ((gchar *)ldtp_req->action_name);
	if (command == LDTP_CMD_INVALID) {
		*err = ldtp_error_get_error (LDTP_ERROR_INVALID_COMMAND);
		return;
	}
	switch (command) {
	case LDTP_CMD_SETAPPMAP:
	case LDTP_CMD_INITAPPMAP: {
		char *appmap_filename = NULL;
		appmap_filename = g_slist_nth_data (cctxt->req->arg_list, 0);
		if (!appmap_filename) {
			*err = ldtp_error_get_error (LDTP_ERROR_OPENING_APPMAP_FILE);
			return;
		}
		g_print ("Appmap file: %s\n", appmap_filename);
		if (cctxt->app_map) {
			ldtp_appmap_free (cctxt->app_map);
			cctxt->app_map = NULL;
		}
		cctxt->app_map = appmap_init (appmap_filename, cctxt->log_fp);
		if (!cctxt->app_map)
			*err = ldtp_error_get_error (LDTP_ERROR_OPENING_APPMAP_FILE);
		return;
	}

	case LDTP_CMD_LOG: {
		char *log;
		char *mode = "";
		log = (char *)cctxt->req->context;
		if (cctxt->req->component)
			mode = (char *)cctxt->req->component;

		/*
		  If startlog is called only once, let us try to utilize
		  tmp_fp for the rest of connections
		*/
		if (cctxt->log_fp == NULL)
			cctxt->log_fp = tmp_fp;

		if (g_ascii_strcasecmp (mode, "PASS") == 0)
			log_msg (LDTP_LOG_PASS, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "FAIL") == 0)
			log_msg (LDTP_LOG_FAIL, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "ERROR") == 0)
			log_msg (LDTP_LOG_ERROR, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "CAUSE") == 0)
			log_msg (LDTP_LOG_CAUSE, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "COMMENT") == 0)
			log_msg (LDTP_LOG_COMMENT, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "INFO") == 0)
			log_msg (LDTP_LOG_INFO, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "MEMINFO") == 0)
			log_msg (LDTP_LOG_MEMINFO, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "CPUINFO") == 0)
			log_msg (LDTP_LOG_CPUINFO, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "GROUPSTART") == 0)
			log_msg (LDTP_LOG_GROUP_START, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "GROUPSTATUS") == 0)
			log_msg (LDTP_LOG_GROUP_STATUS, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "GROUPEND") == 0)
			log_msg (LDTP_LOG_GROUP_END, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "SCRIPTSTART") == 0)
			log_msg (LDTP_LOG_SCRIPT_START, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "SCRIPTEND") == 0)
			log_msg (LDTP_LOG_SCRIPT_END, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "DATAFILENAME") == 0)
			log_msg (LDTP_LOG_DATA_FILENAME, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "WARNING") == 0)
			log_msg (LDTP_LOG_WARNING, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "TESTSTART") == 0)
			log_msg (LDTP_LOG_TESTSTART, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "TESTCASEID") == 0)
			log_msg (LDTP_LOG_TESTCASEID, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "TESTEND") == 0)
			log_msg (LDTP_LOG_TESTEND, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "BEGIN") == 0)
			log_msg (LDTP_LOG_BEGIN, log, cctxt->log_fp);
		else if (g_ascii_strcasecmp (mode, "END") == 0)
			log_msg (LDTP_LOG_END, log, cctxt->log_fp);
		else
			log_msg (LDTP_LOG_DEBUG, log, cctxt->log_fp); // Default is DEBUG mode
		return;
	}

	case LDTP_CMD_START_LOG: {
		int overwrite = 1;
		if (cctxt->req->component)
			overwrite = atol ((char *)cctxt->req->component);
		// Open log file
		if (open_log_file ((char *)cctxt->req->context,
				   overwrite == 0 ? LDTP_LOG_FILE_APPEND : LDTP_LOG_FILE_DELETE,
				   &cctxt->log_fp))
			*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
		else
			*err = ldtp_error_get_error (LDTP_ERROR_OPENING_LOG_FILE);
		tmp_fp = cctxt->log_fp;
		return;
	}

	case LDTP_CMD_STOP_LOG:
		close_log_file (&cctxt->log_fp);
		cctxt->log_fp = NULL;
		return;

	case LDTP_CMD_STOP_SCRIPT_ENGINE:
		*err = ldtp_error_get_error (LDTP_ERROR_STOP_SCRIPT_ENGINE);
		return;

	case LDTP_CMD_SETCONTEXT:
		set_new_context ((char *)cctxt->req->context, (char *)cctxt->req->component);
		return;

	case LDTP_CMD_RELEASECONTEXT:
		release_last_context ();
		return;

	case LDTP_CMD_REINITLDTP: {
		int leaked;
		int  init_error;

		if (cctxt->app_handle) {
			Accessible_unref (cctxt->app_handle);
			cctxt->app_handle = NULL;
		}

		if ((leaked = SPI_exit ())) {
			printf ("Leaked %d SPI handles\n", leaked);
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_REINIT_LDTP);
			return;
		}

		init_error = SPI_init ();

		if (init_error) {
			printf ("Error: SPI Init\n");
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_REINIT_LDTP);
			return;
		}
		*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
		return;
	}

	case LDTP_CMD_GUIEXIST:
		ldtp_gui_gui_exist (cctxt, err);
		return;

	case LDTP_CMD_WAITTILLGUIEXIST:
		ldtp_gui_wait_till_gui_exist (cctxt, err);
		return;

	case LDTP_CMD_WAITTILLGUINOTEXIST:
		ldtp_gui_wait_till_gui_not_exist (cctxt, err);
		return;

	case LDTP_CMD_DOESMENUITEMEXIST:
	case LDTP_CMD_MENUCHECK:
	case LDTP_CMD_MENUUNCHECK:
	case LDTP_CMD_VERIFYMENUCHECK:
	case LDTP_CMD_VERIFYMENUUNCHECK:
	case LDTP_CMD_SELECTMENUITEM:
		if (g_utf8_strchr ((gchar *)ldtp_req->component, -1, ';')) {
			char *token = NULL;
			char *rest_token = NULL;
			char *rest_params = NULL;

			token = strtok (g_strdup ((gchar *)ldtp_req->component), ";");
			if (!token)
				break;
			
			g_free (ldtp_req->component);
			ldtp_req->component = g_strdup (token);
			rest_token = strtok (NULL, ";");
			while (rest_token) {
				if (!rest_params)
					rest_params  = g_strdup (rest_token);
				else {
					rest_params = (char *) realloc (rest_params,
									sizeof (char) *
									(strlen (rest_params) +
									 strlen (rest_token) + 1)
									+ 1);
					strcat (rest_params, ";");
					strcat (rest_params, rest_token);
				}
				rest_token = strtok (NULL, ";");
			}
			cctxt->req->arg_list = g_slist_append (cctxt->req->arg_list, 
							       rest_params);
		}
		break;
	case LDTP_CMD_ONWINDOWCREATE:
		register_window_creation_event (cctxt, err);
		return;
	case LDTP_CMD_REMAP:
		update_cur_window_appmap_handle (cctxt, err);
		return;
	case LDTP_CMD_BINDTEXT: {
		gchar *mode = NULL;
  
		if (cctxt->req->arg_list)
			mode = g_slist_nth_data (cctxt->req->arg_list, 0);

		bind_text (cctxt, mode, err);
		return;
	}
	case LDTP_CMD_GETWINDOWLIST: {
		if (cctxt->app_map) {
			char *data = NULL;
			/*
			  If appliction map is not initialized for new windows,
			  then call to update_window_list will get those windows too
			  NOTE: update_window_list will fail if none of the applications
			  has been accessed, ie., atleast one component function has
			  to be called in one of the current running application 
			*/
			/*
			  FIXME: Should be implemented ;)
			  // update_window_list (cctxt, err);
			*/
			g_hash_table_foreach (cctxt->app_map, &add_item_to_list, (gpointer)&data);
			cctxt->resp->data = data;
			if (cctxt->resp->data) {
				char *tmp = NULL;
				/*
				  FRAME the output in XML format
				*/
				tmp = g_strdup_printf ("%s<OBJECTLIST>%s</OBJECTLIST>", XML_HEADER, data);
				if (tmp) {
					g_free (data);
					cctxt->resp->data = tmp;
					g_print ("LIST: %s\n", cctxt->resp->data);
					cctxt->resp->data_len = g_utf8_strlen (cctxt->resp->data, -1);
				}
				else {
					*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
					g_print ("%s\n", (*err)->message);
					log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
				}
			}
			else {
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
				g_print ("%s\n", (*err)->message);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			}
		}
		else {
			*err = ldtp_error_get_error (LDTP_ERROR_APPMAP_NOT_INITIALIZED);
			g_print ("%s - %s\n", __FILE__, (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
		}
		return;
	}

	case LDTP_CMD_GETOBJECTLIST: {
		char *data = NULL;
		GHashTable *ht = NULL;

		if (!cctxt->req->context) {
			*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		if (!cctxt->app_map) {
                        /* if appmap not already initialised, do it now */
                        accessible = ldtp_gui_get_gui_handle (cctxt, err);
                        if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
                                return;
                        }
                        ldtp_gui_free_gui_handle (accessible);
		}
		ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
		if (!ht) {
			update_cur_window_appmap_handle (cctxt,
							 err);
			ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
			if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				g_print ("Unable to update context: %s in appmap", (char *)cctxt->req->context);
				return;
			}
		}
		if (!ht) {
			g_print ("Unable to find context\n");
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_CONTEXT_HANDLE);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		g_hash_table_foreach (ht, &add_item_to_list, (gpointer)&data);
		cctxt->resp->data = data;
		if (cctxt->resp->data) {
			char *tmp = NULL;
			/*
			  FRAME the output in XML format
			*/
			tmp = g_strdup_printf ("%s<OBJECTLIST>%s</OBJECTLIST>", XML_HEADER, data);
			if (tmp) {
				g_free (data);
				cctxt->resp->data = tmp;
				g_print ("LIST: %s\n", cctxt->resp->data);
				cctxt->resp->data_len = g_utf8_strlen (cctxt->resp->data, -1);
			}
			else {
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
				g_print ("%s\n", (*err)->message);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			}
		}
		else {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
		}
		return;
	}

	case LDTP_CMD_GETOBJECTINFO: {
		char *data = NULL;
		GHashTable *ht = NULL;

		if (!cctxt->req->context || !cctxt->req->component) {
			*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		if (!cctxt->app_map) {
                        /* if appmap not already initialised, do it now */
                        accessible = ldtp_gui_get_gui_handle (cctxt, err);
                        if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
                                return;
                        }
                        ldtp_gui_free_gui_handle (accessible);
		}
		ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
		if (!ht) {
			update_cur_window_appmap_handle (cctxt,
							 err);
			ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
			if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				g_print ("Unable to update context: %s in appmap", (char *)cctxt->req->context);
				return;
			}
		}
		if (!ht) {
			g_print ("Unable to find context\n");
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_CONTEXT_HANDLE);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		ht = get_object_def (ht, (char *)cctxt->req->component, cctxt->log_fp);
		if (!ht) {
			update_cur_window_appmap_handle (cctxt,
							 err);
			ht = get_object_def (ht, (char *)cctxt->req->component, cctxt->log_fp);
			if (!ht) {
				g_print ("Unable to find component\n");
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_COMPONENT_HANDLE);
				g_print ("%s\n", (*err)->message);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
				return;
			}
		}
		g_hash_table_foreach (ht, &add_item_to_list, (gpointer)&data);
		cctxt->resp->data = data;
		if (cctxt->resp->data) {
			char *tmp = NULL;
			/*
			  FRAME the output in XML format
			*/
			tmp = g_strdup_printf ("%s<OBJECTLIST>%s</OBJECTLIST>", XML_HEADER, data);
			if (tmp) {
				g_free (data);
				cctxt->resp->data = tmp;
				g_print ("LIST: %s\n", cctxt->resp->data);
				cctxt->resp->data_len = g_utf8_strlen (cctxt->resp->data, -1);
			}
			else {
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
				g_print ("%s\n", (*err)->message);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			}
		}
		else {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_OBJECT_LIST);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
		}
		return;
	}

	case LDTP_CMD_GETOBJECTPROPERTY: {
		char *property = NULL;
		char *obj_property = NULL;
		GHashTable *ht = NULL;

		if (!cctxt->req->context || !cctxt->req->component) {
			*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		if (!cctxt->app_map) {
                        /* if appmap not already initialised, do it now */
                        accessible = ldtp_gui_get_gui_handle (cctxt, err);
                        if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
                                return;
                        }
                        ldtp_gui_free_gui_handle (accessible);
		}
		ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
		if (!ht) {
			update_cur_window_appmap_handle (cctxt,
							 err);
			ht = get_object_def (cctxt->app_map, (char *)cctxt->req->context, cctxt->log_fp);
			if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
				g_print ("Unable to update context: %s in appmap", (char *)cctxt->req->context);
				return;
			}
		}
		if (!ht) {
			g_print ("Unable to find context\n");
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_CONTEXT_HANDLE);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		ht = get_object_def (ht, (char *)cctxt->req->component, cctxt->log_fp);
		if (!ht) {
			update_cur_window_appmap_handle (cctxt,
							 err);
			ht = get_object_def (ht, (char *)cctxt->req->component, cctxt->log_fp);
			if (!ht) {
				g_print ("Unable to find component\n");
				*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_COMPONENT_HANDLE);
				g_print ("%s\n", (*err)->message);
				log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
				return;
			}
		}
		property = g_slist_nth_data (cctxt->req->arg_list, 0);
		obj_property = get_property (ht, property, cctxt->log_fp);
		if (!obj_property) {
			g_print ("Unable to find property\n");
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_PROPERTY);
			g_print ("%s\n", (*err)->message);
			log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
			return;
		}
		cctxt->resp->data = g_strdup (obj_property);
		cctxt->resp->data_len = g_utf8_strlen (obj_property, -1);
		return;
	}
	
	case LDTP_CMD_STARTRECORD: {
		g_print ("Entered LDTP_CMD_STARTRECORD\n");
		record_main (cctxt, command);
		return ;
	}
	case LDTP_CMD_STOPRECORD: {
		g_print ("Entered LDTP_CMD_STOPRECORD\n");
		record_main (cctxt,command);
		return ;
	}
	}

	/* 
	   1)  Get the Accessible object corresponding to "component-name".
	   2)  Get the class-id and
	   3)  Call the corresponding component main function with necessary
	   arguments.
	*/

	accessible = ldtp_gui_get_gui_handle (cctxt, err);
	if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
		/* FIXME: Error handling */
		return;
	}

	/* Update the gui handle in the context */
	cctxt->gui_handle = accessible;

	if (command == LDTP_CMD_HASSTATE) {
		has_state (cctxt, err);
		return;
	}

	switch (cctxt->gui_handle->class_id) {
	case SPI_ROLE_COMBO_BOX:
		*err = combo_box_main (cctxt, command);
		break;

	case SPI_ROLE_CHECK_BOX:
		*err = check_box_main (cctxt, command);
		break;

	case SPI_ROLE_CHECK_MENU_ITEM:
		*err = check_menu_item_main (cctxt, command);
		break;

	case SPI_ROLE_RADIO_BUTTON:
		*err = radio_button_main (cctxt, command);
		break;

	case SPI_ROLE_RADIO_MENU_ITEM:
		*err = radio_menu_item_main (cctxt, command);
		break;

	case SPI_ROLE_PUSH_BUTTON:
		*err = push_button_main (cctxt, command);
		break;

	case SPI_ROLE_TOGGLE_BUTTON:
		*err = toggle_button_main (cctxt, command);
		break;

	case SPI_ROLE_LIST:
		*err = list_main (cctxt, command);
		break;

	case SPI_ROLE_TEXT:
	case SPI_ROLE_EDITBAR:
	case SPI_ROLE_PASSWORD_TEXT:
		*err = text_main (cctxt, command);
		break;

	case SPI_ROLE_PANEL:
		*err = panel_main (cctxt, command);
		break;

	case SPI_ROLE_ICON:
		*err = icon_main (cctxt, command);
		break;

	case SPI_ROLE_LAYERED_PANE:
		*err = layered_pane_main (cctxt, command);
		break;

	case SPI_ROLE_LABEL:
		*err = label_main (cctxt, command);
		break;

	case SPI_ROLE_MENU:
		*err = menu_main (cctxt, command,
				  (gchar *)ldtp_req->context);
		break;

	case SPI_ROLE_PAGE_TAB_LIST:
		*err = page_tab_list_main (cctxt, command);
		break;

	case CALENDAR_VIEW:
		*err = calendar_view_main (cctxt, command);
		break;

	case SPI_ROLE_CALENDAR:
		*err = calendar_main (cctxt, command);
		break;

	case SPI_ROLE_SPIN_BUTTON:
		*err = spin_button_main (cctxt, command);
		break;

	case SPI_ROLE_TABLE:
		*err = table_main (cctxt, command);
		break;

	case SPI_ROLE_TREE_TABLE:
		*err = tree_table_main (cctxt, command);
		break;

	case SPI_ROLE_SCROLL_BAR:
		*err = scroll_bar_main (cctxt, command);
		break;

	case SPI_ROLE_SLIDER:
		*err = slider_main (cctxt, command);
		break;

	case SPI_ROLE_STATUS_BAR:
		*err = status_bar_main (cctxt, command);
		break;

	case SPI_ROLE_TOOL_BAR:
		*err = tool_bar_main (cctxt, command);
		break;
	default:
		*err = ldtp_error_get_error (LDTP_ERROR_ROLE_NOT_IMPLEMENTED);
		break;
	}

	/* Need to update cctxt->resp with appropriate status_code and data.
	   For this, we have to have a switch case on command-id, because
	   not all commands return data.
	*/
	ldtp_gui_free_gui_handle (accessible);
	cctxt->gui_handle = NULL;
}

void 
handle_client (void *ptr)
{
	int* sock_fd = (int *)ptr;
	char* packet = NULL;
	char* tmpptr = NULL;
	uint32_t packet_len;
	uint32_t pckt_len = 0;
	uint32_t prev_pckt_len = 0;
	size_t i = 0;
	size_t bytes_read = 0;
	uint32_t resp_size = 0;
	char *resp_pckt = NULL;

	LDTPError* err = NULL;
	gchar *locale_lang = NULL;

	LDTPClientContext *cctxt = NULL;
	Packet *pckt = NULL;

	cctxt = calloc (1, sizeof (LDTPClientContext));
	cctxt->app_map = NULL;
	cctxt->req = calloc (1, sizeof (LDTPRequest));
	ldtp_request_init (cctxt->req);
	cctxt->resp = calloc (1, sizeof (LDTPResponse));
	cctxt->sock_fd = *sock_fd;
	cctxt->log_fp = NULL;

	/*
	  Check whether locale lang set or not
	*/
	locale_lang = getenv ("LANG");
	if (locale_lang && 
	    g_utf8_collate (locale_lang, "en_US.UTF-8") != 0) {
		g_print ("Locale languages: %s\n", locale_lang);
		cctxt->locale_set = TRUE;
		setlocale (LC_ALL, "");
	} else
		cctxt->locale_set = FALSE;

	while (1) {
		packet_len = 0;
		ldtp_read_sizet (*sock_fd, &packet_len, &err);
		if (err && err->err_code != LDTP_ERROR_SUCCESS) {
			log_msg (LDTP_LOG_CAUSE, err->message, cctxt->log_fp);
			goto error;
		}

		packet_len = ntohl (packet_len);
		g_print ("Client packet len: %d\n", packet_len);

		if (packet_len <= 0) {
			err = ldtp_error_get_error (LDTP_ERROR_PACKET_INVALID);
			log_msg (LDTP_LOG_CAUSE, err->message, cctxt->log_fp);
			goto error;
		}

		if (packet_len > prev_pckt_len)
			prev_pckt_len = packet_len;
		packet = ldtp_memory_do_alloc (packet, packet_len, 
					       prev_pckt_len);
		tmpptr = packet;

		pckt_len = packet_len;
		while (pckt_len > 0) {
			g_print ("i = %d\n", i);
			pckt_len = pckt_len > 512 ? 512 : pckt_len;
			ldtp_read_data (*sock_fd, packet, pckt_len, 
					&bytes_read, &err);
			if (err && err->err_code != LDTP_ERROR_SUCCESS) {
				/*
				  FIXME:
				  ldtp_log ("Connection: closed [%s:%d]\n", 
				  client->ip_address, client->socketfd);
				*/
				g_print ("Client data read error\n");
				log_msg (LDTP_LOG_CAUSE, err->message, cctxt->log_fp);
				goto error;
			}
			if (packet)
				g_print ("Data read %d, packet-len = %d, bytes read = %d, data: %s\n", 
					 pckt_len, packet_len, bytes_read, packet);
			packet += bytes_read;
			pckt_len = packet_len - (packet - tmpptr);
			g_print ("PACKET LENGTH: %d\n", pckt_len);
			if (pckt_len <= 0 || bytes_read == 0)
				break;
		}
		packet = tmpptr;

		g_print ("Received packet [%s] through %d\n", packet, *sock_fd);

		pckt = calloc (1, sizeof (Packet));
		pckt->packet = packet;
		pckt->len = packet_len;

		handle_request (cctxt, pckt, &err);
		if (err && err->err_code != LDTP_ERROR_SUCCESS) {
			/*
			  FIXME:
			  ldtp_log ("Protocol error: closing connection [%s:%d]\n", 
			  client->ip_address, client->socketfd);
			*/
			/*
			  We should not close the connection for all errors
			*/
			//goto error;
		}
		if (err && err->err_code == LDTP_ERROR_STOP_SCRIPT_ENGINE) {
			goto error;
		}
		if (cctxt->req->arg_list && cctxt->req->arg_list->data) {
			cctxt->req->arg_list = g_slist_remove_all (cctxt->req->arg_list,
								   cctxt->req->arg_list->data);
		}
		if (packet)
			g_free (packet);
		packet = NULL;

		if (pckt)
			g_free (pckt);
		pckt = NULL;

		cctxt->resp->resp_status = err;
		err = NULL;
		generate_response_packet (cctxt, &err, &resp_pckt, &resp_size);
		if (err && err->err_code != LDTP_ERROR_SUCCESS) {
			log_msg (LDTP_LOG_CAUSE, err->message, cctxt->log_fp);
			g_print ("Error generating response\n");
			goto error;
		}
		send_response (cctxt->sock_fd, resp_pckt, resp_size, &err);
		g_free (resp_pckt);
		/* FIXME:
		   Error handling;
		*/
		if (cctxt->resp->data) {
			g_free (cctxt->resp->data);
			cctxt->resp->data = NULL;
		}
		cctxt->resp->data_len = 0;

		/*
		  FIXME:
		  if (args->verbose || args->daemon_mode)
		  ldtp_log ("Received packet: [%s:%d]\n[%s]\n", 
		  client->ip_address, client->socketfd, 
		  client->message);
		*/
	}
	pthread_exit ((void *)0);

 error:
	g_print ("handle_client: error:\n");
	cctxt->resp->resp_status = err;
	err = NULL;
	generate_response_packet (cctxt, &err, &resp_pckt, &resp_size);
	if (err && err->err_code != LDTP_ERROR_SUCCESS) {
		log_msg (LDTP_LOG_CAUSE, err->message, cctxt->log_fp);
		g_print ("Error generating response\n");
	}
	else {
		send_response (cctxt->sock_fd, resp_pckt, resp_size, &err);
		g_free (resp_pckt);
	}
	close_connection (*sock_fd);
	if (packet)
		g_free (packet);

	if (sock_fd)
		g_free (sock_fd);

	if (pckt)
		g_free (pckt);
	if (cctxt->app_handle) {
		/*	
		  FIXME: If the server closes connection, before the client closes connection,
		  then we get 1 SPI handle leak. Should be handled.
		*/
		Accessible_unref (cctxt->app_handle);
		cctxt->app_handle = NULL;
	}

	ldtp_appmap_free (cctxt->app_map);
	err = cctxt->resp->resp_status;
	if (err && err->err_code == LDTP_ERROR_STOP_SCRIPT_ENGINE) {
		cleanup (0);
	}
	pthread_exit ((void *) 1);
}
