/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Veerapuram Varadhan <v.varadhan@gmail.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <glib.h>

#include "ldtp-command.h"

static GHashTable* cmd_id_table = NULL;
 
static LDTPCommand*
ldtp_command_new (LDTPCommandId id, const char* command)
{
	LDTPCommand* cmd = calloc (1, sizeof (LDTPCommand));
	cmd->cmd_id = id;
	cmd->cmd_name = g_strdup (command);
	return cmd;
}

static void
ldtp_command_free (LDTPCommand* cmd)
{
	if (!cmd)
		return;

	g_free (cmd->cmd_name);
	g_free (cmd);
}

static void
ldtp_command_init_command_table ()
{
	int i = 0;

	LDTPCommand *command[] = 
		{
			ldtp_command_new ( LDTP_CMD_LOG, "log" ),
			ldtp_command_new ( LDTP_CMD_START_LOG, "startlog" ),
			ldtp_command_new ( LDTP_CMD_STOP_LOG, "stoplog" ),
			ldtp_command_new ( LDTP_CMD_CHECK, "check" ),
			ldtp_command_new ( LDTP_CMD_UNCHECK, "uncheck" ),
			ldtp_command_new ( LDTP_CMD_MENUCHECK, "menucheck" ),
			ldtp_command_new ( LDTP_CMD_MENUUNCHECK, "menuuncheck" ),
			ldtp_command_new ( LDTP_CMD_CLICK, "click" ),
			ldtp_command_new ( LDTP_CMD_VERIFYCHECK, "verifycheck" ),
			ldtp_command_new ( LDTP_CMD_VERIFYUNCHECK, "verifyuncheck" ),
			ldtp_command_new ( LDTP_CMD_VERIFYMENUCHECK, "verifymenucheck" ),
			ldtp_command_new ( LDTP_CMD_VERIFYMENUUNCHECK, "verifymenuuncheck" ),
			ldtp_command_new ( LDTP_CMD_HIDELIST, "hidelist" ),
			ldtp_command_new ( LDTP_CMD_COMBOSELECTINDEX, "comboselectindex" ),
			ldtp_command_new ( LDTP_CMD_SELECTINDEX, "selectindex" ),
			ldtp_command_new ( LDTP_CMD_SETTEXTVALUE, "settextvalue" ),
			ldtp_command_new ( LDTP_CMD_GETTEXTVALUE, "gettextvalue" ),
			ldtp_command_new ( LDTP_CMD_APPENDTEXT, "appendtext" ),
			ldtp_command_new ( LDTP_CMD_ACTIVATETEXT, "activatetext" ),
			ldtp_command_new ( LDTP_CMD_CUTTEXT, "cuttext" ),
			ldtp_command_new ( LDTP_CMD_PASTETEXT, "pastetext" ),
			ldtp_command_new ( LDTP_CMD_DELETETEXT, "deletetext" ),
			ldtp_command_new ( LDTP_CMD_SELECTTEXTBYNAME, "selecttextbyname" ),
			ldtp_command_new ( LDTP_CMD_SELECTTEXTBYINDEXANDREGION, "selecttextbyindexandregion" ),
			ldtp_command_new ( LDTP_CMD_SHOWLIST, "showlist" ),
			ldtp_command_new ( LDTP_CMD_VERIFYDROPDOWN, "verifydropdown" ),
			ldtp_command_new ( LDTP_CMD_VERIFYHIDELIST, "verifyhidelist" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSHOWLIST, "verifyshowlist" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSELECT, "verifyselect" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSETTEXT, "verifysettext" ),
			ldtp_command_new ( LDTP_CMD_RIGHTCLICK, "rightclick" ),
			ldtp_command_new ( LDTP_CMD_GETLABEL, "getlabel" ),
			ldtp_command_new ( LDTP_CMD_GETLABELATINDEX, "getlabelatindex" ),
			ldtp_command_new ( LDTP_CMD_SELECTPOPUPMENU, "selectpopupmenu" ),
			ldtp_command_new ( LDTP_CMD_SELECTITEM, "selectitem" ),
			ldtp_command_new ( LDTP_CMD_COMBOSELECT, "comboselect" ),
			ldtp_command_new ( LDTP_CMD_SELECTTEXTITEM, "selecttextitem" ),
			ldtp_command_new ( LDTP_CMD_ISMENUITEMENABLED, "ismenuitemenabled" ),
			ldtp_command_new ( LDTP_CMD_SELECTMENUITEM, "selectmenuitem" ),
			ldtp_command_new ( LDTP_CMD_SELECTTAB, "selecttab" ),
			ldtp_command_new ( LDTP_CMD_VERIFYPUSHBUTTON, "verifypushbutton" ),
			ldtp_command_new ( LDTP_CMD_STATEENABLED, "stateenabled" ),
			ldtp_command_new ( LDTP_CMD_SETVALUE, "setvalue" ),
			ldtp_command_new ( LDTP_CMD_GETVALUE, "getvalue" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSETVALUE, "verifysetvalue" ),
			ldtp_command_new ( LDTP_CMD_SELECTROW, "selectrow" ),
			ldtp_command_new ( LDTP_CMD_SELECTROWINDEX, "selectrowindex" ),
			ldtp_command_new ( LDTP_CMD_SELECTROWPARTIALMATCH, "selectrowpartialmatch" ),
			ldtp_command_new ( LDTP_CMD_VERIFYTOGGLED, "verifytoggled" ),
			ldtp_command_new ( LDTP_CMD_ONEDOWN, "onedown" ),
			ldtp_command_new ( LDTP_CMD_ONEUP, "oneup" ),
			ldtp_command_new ( LDTP_CMD_ONERIGHT, "oneright" ),
			ldtp_command_new ( LDTP_CMD_ONELEFT, "oneleft" ),
			ldtp_command_new ( LDTP_CMD_SCROLLDOWN,  "scrolldown" ),
			ldtp_command_new ( LDTP_CMD_SCROLLUP, "scrollup" ),
			ldtp_command_new ( LDTP_CMD_SCROLLRIGHT, "scrollright" ),
			ldtp_command_new ( LDTP_CMD_SCROLLLEFT, "scrollleft" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSCROLLBAR, "verifyscrollbar" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSCROLLBARVERTICAL, "verifyscrollbarvertical" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSCROLLBARHORIZONTAL, "verifyscrollbarhorizontal" ),
			ldtp_command_new ( LDTP_CMD_SETMAX, "setmax" ),
			ldtp_command_new ( LDTP_CMD_SETMIN, "setmin" ),
			ldtp_command_new ( LDTP_CMD_INCREASE, "increase" ),
			ldtp_command_new ( LDTP_CMD_DECREASE, "decrease" ),
			ldtp_command_new ( LDTP_CMD_SELECTPANEL, "selectpanel" ),
			ldtp_command_new ( LDTP_CMD_SELECTPANELNAME, "selectpanelname" ),
			ldtp_command_new ( LDTP_CMD_GETPANELCHILDCOUNT, "getpanelchildcount" ),
			ldtp_command_new ( LDTP_CMD_VERIFYTABLECELL,  "verifytablecell" ),
			ldtp_command_new ( LDTP_CMD_SETCELLVALUE, "setcellvalue" ),
			ldtp_command_new ( LDTP_CMD_GETCELLVALUE, "getcellvalue" ),
			ldtp_command_new ( LDTP_CMD_SELECTLASTROW, "selectlastrow" ),
			ldtp_command_new ( LDTP_CMD_SELECTLABELSPANELBYNAME, "selectlabelspanelbyname" ),
			ldtp_command_new ( LDTP_CMD_SETCONTEXT, "setcontext" ),
			ldtp_command_new ( LDTP_CMD_RELEASECONTEXT, "releasecontext" ),
			ldtp_command_new ( LDTP_CMD_GETSTATUSBARTEXT, "getstatusbartext" ),
			ldtp_command_new ( LDTP_CMD_SELECTEVENT, "selectevent" ),
			ldtp_command_new ( LDTP_CMD_SELECTEVENTINDEX, "selecteventindex" ),
			ldtp_command_new ( LDTP_CMD_DOESROWEXIST, "doesrowexist" ),
			ldtp_command_new ( LDTP_CMD_DOESMENUITEMEXIST, "doesmenuitemexist" ),
			ldtp_command_new ( LDTP_CMD_LISTSUBMENUS, "listsubmenus" ),
			ldtp_command_new ( LDTP_CMD_CHECKROW, "checkrow" ),
			ldtp_command_new ( LDTP_CMD_VERIFYPARTIALMATCH, "verifypartialmatch" ),
			ldtp_command_new ( LDTP_CMD_GETROWCOUNT, "getrowcount" ),
			ldtp_command_new ( LDTP_CMD_VERIFYPARTIALTABLECELL, "verifypartialtablecell" ),
			ldtp_command_new ( LDTP_CMD_GRABFOCUS, "grabfocus" ),
			ldtp_command_new ( LDTP_CMD_VERIFYEVENTEXIST, "verifyeventexist" ),
			ldtp_command_new ( LDTP_CMD_EXPANDTABLECELL, "expandtablecell" ),
			ldtp_command_new ( LDTP_CMD_GETTREETABLEROWINDEX, "gettreetablerowindex" ),
			ldtp_command_new ( LDTP_CMD_VERIFYCHECKROW, "verifycheckrow" ),
			ldtp_command_new ( LDTP_CMD_VERIFYUNCHECKROW, "verifyuncheckrow" ),
			ldtp_command_new ( LDTP_CMD_VERIFYVISIBLEBUTTONCOUNT, "verifyvisiblebuttoncount" ),
			ldtp_command_new ( LDTP_CMD_VERIFYBUTTONCOUNT, "verifybuttoncount" ),
			ldtp_command_new ( LDTP_CMD_GETEVENTCOUNT, "geteventcount" ),
			ldtp_command_new ( LDTP_CMD_DOUBLECLICKROW, "doubleclickrow" ),
			ldtp_command_new ( LDTP_CMD_GETTABLEROWINDEX, "gettablerowindex" ),
			ldtp_command_new ( LDTP_CMD_GETCHARCOUNT, "getcharcount" ),
			ldtp_command_new ( LDTP_CMD_ISTEXTSTATEENABLED, "istextstateenabled" ),
			ldtp_command_new ( LDTP_CMD_COPYTEXT, "copytext" ),
			ldtp_command_new ( LDTP_CMD_INSERTTEXT, "inserttext" ),
			ldtp_command_new ( LDTP_CMD_UNCHECKROW, "uncheckrow" ),
			ldtp_command_new ( LDTP_CMD_CAPTURETOFILE, "capturetofile" ),
			ldtp_command_new ( LDTP_CMD_DOUBLECLICK, "doubleclick" ),
			ldtp_command_new ( LDTP_CMD_GETTEXTPROPERTY, "gettextproperty" ),
			ldtp_command_new ( LDTP_CMD_COMPARETEXTPROPERTY, "comparetextproperty" ),
			ldtp_command_new ( LDTP_CMD_CONTAINSTEXTPROPERTY, "containstextproperty" ),
			ldtp_command_new ( LDTP_CMD_SELECTCALENDARDATE, "selectcalendardate" ),
			ldtp_command_new ( LDTP_CMD_SELECTTABINDEX , "selecttabindex" ),
			ldtp_command_new ( LDTP_CMD_SORTCOLUMNINDEX , "sortcolumnindex" ),
			ldtp_command_new ( LDTP_CMD_SORTCOLUMN, "sortcolumn" ),
			ldtp_command_new ( LDTP_CMD_REINITLDTP, "reinitldtp" ),
			ldtp_command_new ( LDTP_CMD_INITAPPMAP, "initappmap" ),
			ldtp_command_new ( LDTP_CMD_SETAPPMAP, "setappmap" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSLIDER, "verifyslider" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSLIDERVERTICAL, "verifyslidervertical" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSLIDERHORIZONTAL, "verifysliderhorizontal" ),
			ldtp_command_new ( LDTP_CMD_GETSLIDERVALUE, "getslidervalue" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSTATUSBAR, "verifystatusbar" ),
			ldtp_command_new ( LDTP_CMD_VERIFYSTATUSBARVISIBLE, "verifystatusbarvisible" ),
			ldtp_command_new ( LDTP_CMD_WAITTILLGUIEXIST, "waittillguiexist" ),
			ldtp_command_new ( LDTP_CMD_WAITTILLGUINOTEXIST, "waittillguinotexist" ),
			ldtp_command_new ( LDTP_CMD_GUIEXIST, "guiexist" ),
			ldtp_command_new ( LDTP_CMD_ONWINDOWCREATE, "onwindowcreate" ),
			ldtp_command_new ( LDTP_CMD_REMAP, "remap" ),
			ldtp_command_new ( LDTP_CMD_GETWINDOWLIST, "getwindowlist"),
			ldtp_command_new ( LDTP_CMD_GETOBJECTLIST, "getobjectlist"),
			ldtp_command_new ( LDTP_CMD_GETOBJECTINFO, "getobjectinfo"),
			ldtp_command_new ( LDTP_CMD_GETOBJECTPROPERTY, "getobjectproperty"),
			ldtp_command_new ( LDTP_CMD_BINDTEXT, "bindtext"),
                        ldtp_command_new ( LDTP_CMD_MOUSE_LEFT_CLICK, "mouseleftclick"),
                        ldtp_command_new ( LDTP_CMD_MOUSE_RIGHT_CLICK, "mouserightclick"),
                        ldtp_command_new ( LDTP_CMD_MOUSE_MOVE, "mousemove"),
			ldtp_command_new ( LDTP_CMD_KBD_ENTER, "enterstring"),
			ldtp_command_new ( LDTP_CMD_SET_CURSOR, "setcursorposition"),
			ldtp_command_new ( LDTP_CMD_GET_CURSOR, "getcursorposition"),
			ldtp_command_new ( LDTP_CMD_GET_TAB_COUNT, "gettabcount"),
			ldtp_command_new ( LDTP_CMD_STOP_SCRIPT_ENGINE, "stopscriptengine"),
			ldtp_command_new ( LDTP_CMD_STARTRECORD, "startrecord"),
			ldtp_command_new ( LDTP_CMD_STOPRECORD, "stoprecord"),
			ldtp_command_new ( LDTP_CMD_HASSTATE, "hasstate"),
			NULL
		};

	cmd_id_table = g_hash_table_new_full (&g_str_hash, &g_str_equal, NULL,  (GDestroyNotify)&ldtp_command_free);
	while (command[i]) {
		g_hash_table_insert (cmd_id_table, command[i]->cmd_name, command[i]); 
		i++;
	}
}

void
ldtp_command_destroy_command_table (void)
{
	g_hash_table_destroy (cmd_id_table);
}

LDTPCommandId
ldtp_command_get_command_id (const char* command_name)
{
	LDTPCommand *cmd;
	if (!command_name)
		return LDTP_CMD_INVALID;
	if (!cmd_id_table)
		ldtp_command_init_command_table ();
	g_print ("Get command id: %s\n", command_name);
	cmd = g_hash_table_lookup (cmd_id_table, command_name);
	return cmd ? cmd->cmd_id : LDTP_CMD_INVALID;
}
