/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Nagappan A <nagappan@gmail.com>
 *    Poornima <pnayak@novell.com>
 *    Premkumar J <jpremkumar@novell.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "ldtp.h"
#include "remap.h"
#include "ldtp-gui.h"
#include "ldtp-state.h"
#include "ldtp-utils.h"
#include "ldtp-logger.h"
#include "localization.h"

static char *last_new_context = NULL;
static char *last_existing_context = NULL;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

object_state state_list[80] = {
	{0, {0}},  // INVALID 0
	{0, {0}},  // ACCEL_LABEL 1
	{5, { SPI_STATE_ENABLED, SPI_STATE_MODAL, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // ALERT 2
	{0, {0}},  // ANIMATION 3
	{0, {0}},  // ARROW 4
	{0, {0}},  // CALENDAR 5
	{0, {0}},  //CANVAS 6
	{2, {SPI_STATE_FOCUSABLE, SPI_STATE_VISIBLE}},  // CHECK_BOX 7
	{3, { SPI_STATE_ENABLED, SPI_STATE_SELECTABLE, SPI_STATE_SENSITIVE}},  // CHECK_MENU_ITEM 8

	{0, {0}},  // COLOR_CHOOSER 9
	{0, {0}},  // COLUMN_HEADER 10

	{4,{SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},  // COMBO_BOX 11

	{0, {0}},  // DATE_EDITOR 12
	{0, {0}},  // DESKTOP_ICON 13
	{4, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // DESKTOP_FRAME 14
	{0, {0}},  // DIAL 15

	{4, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}}, // DIALOG 16

	{0, {0}},  // DIRECTORY_PANE 17
	{0, {0}},  // DRAWING_AREA 18
	{0, {0}},  // FILE_CHOOSER 19

	{5, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VERTICAL, SPI_STATE_VISIBLE	}},  // FILLER 20

	{0, {0}},  // FONT_CHOOSER 21 

	{5, { SPI_STATE_ENABLED, SPI_STATE_RESIZABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // FRAME 22

	{0, {0}},  // GLASS_PANE 23
	{5, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // HTML_CONTAINER 24
	{4, { SPI_STATE_FOCUSABLE, SPI_STATE_SELECTABLE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // ICON 25
	{0, {0}},  // IMAGE 26
	{0, {0}},  // INTERNAL_FRAME 27
	{4, { SPI_STATE_ENABLED, SPI_STATE_MULTI_LINE, SPI_STATE_SENSITIVE, SPI_STATE_VISIBLE }}, // LABEL 28
	{5, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},  // LAYERED_PANE 29

	{6, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_FOCUSABLE, SPI_STATE_SELECTABLE, SPI_STATE_VISIBLE}},  // LIST 30

	{6, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},  //LIST_ITEM 31

	{1, { SPI_STATE_ENABLED}}, //MENU 32

	{4, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},//MENU_BAR 33

	{1, { SPI_STATE_ENABLED }}, // MENU_ITEM 34

	{0, {0}},//OPTION_PANE 35

	{6, {SPI_STATE_ENABLED, SPI_STATE_MULTI_LINE, SPI_STATE_SELECTABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}}, // PAGE_TAB 36

	{4, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},//PAGE_TAB_LIST 37

	{3, {SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_VISIBLE }}, // PANEL 38

	{0, {0}}, //PASSWORD_TEXT 39
	{0, {0}}, //POPUP_MENU 40
	{0, {0}}, //PROGRESS_BAR 41

	{4, { SPI_STATE_ENABLED, SPI_STATE_VISIBLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING}}, //PUSH_BUTTON 42

	{4, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_VISIBLE }},//RADIO_BUTTON 43

	{3, { SPI_STATE_ENABLED, SPI_STATE_SELECTABLE, SPI_STATE_SENSITIVE }},//RADIO_MENU_ITEM 44

	{0, {0}},//ROOT_PANE 45
	{0, {0}},//ROW_HEADER 46
	{2, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE }},//SCROLL_BAR 47

	{5, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},//SCROLL_PANE 48

	{3, { SPI_STATE_ENABLED, SPI_STATE_SELECTABLE, SPI_STATE_SENSITIVE }}, //SEPARATOR 49

	{0, {0}},//SLIDER 50

	{6, { SPI_STATE_EDITABLE, SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SINGLE_LINE, SPI_STATE_VISIBLE }}, //SPIN_BUTTON 51

	{6, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VERTICAL, SPI_STATE_VISIBLE}}, //SPLIT_PANE 52

	{4, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},//STATUS_BAR 53

	{6, {SPI_STATE_VISIBLE, SPI_STATE_SHOWING, SPI_STATE_SENSITIVE, SPI_STATE_MANAGES_DESCENDANTS, SPI_STATE_FOCUSABLE, SPI_STATE_ENABLED	}},//TABLE 54

	{6, { SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SINGLE_LINE, SPI_STATE_TRANSIENT, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }},   //TABLE_CELL 55

	{6, { SPI_STATE_ENABLED, SPI_STATE_SELECTABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_FOCUSABLE, SPI_STATE_VISIBLE}}, //TABLE_COLUMN_HEADER 56

	{0, {0}},//TABLE_ROW_HEADER 57
	{0, {0}},//TEAROFF_MENU_ITEM 58
	{5, { SPI_STATE_ENABLED, SPI_STATE_RESIZABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }}, //TERMINAL 59

	{7, {SPI_STATE_EDITABLE, SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_MULTI_LINE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE }}, //TEXT 60

	{5, {	SPI_STATE_ENABLED, SPI_STATE_FOCUSABLE, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},//TOGGLE_BUTTON 61

	{0, {SPI_STATE_ENABLED,SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}},//TOOL_BAR 62
	{0, {0}},//TOOL_TIP 63
	{0, {0}},//TREE 64
	{0, {0}},//TREE_TABLE 65

	{5, { SPI_STATE_ENABLED, SPI_STATE_HORIZONTAL, SPI_STATE_SENSITIVE, SPI_STATE_SHOWING, SPI_STATE_VISIBLE}}, //UNKNOWN 66

	{2, { SPI_STATE_ENABLED, SPI_STATE_SENSITIVE }},//VIEWPORT 67
	{0, {0}},//WINDOW 68
	{0, {0}},//EXTENDED 69
	{0, {0}},//HEADER 70
	{0, {0}},//FOOTER 71
	{0, {0}},//PARAGRAPH 72
	{0, {0}},//RULER 73
	{0, {0}},//APPLICATION 74
	{0, {0}},//AUTOCOMPLETE 75
	{0, {0}},//EDITBAR 76
	{0, {0}},//LAST_DEFINED 77
	{0, {0}},//CALENDAR_VIEW 78
	{0, {0}},//CALENDAR_EVENT 79
};			

static int
push (struct node *head, int value)
{
	struct node *new_node;

	new_node = (struct node *) malloc (sizeof (struct node));
	if (new_node) {
		new_node->child_index = value;
		new_node->next = head->next;
		head->next = new_node;
		return 1;
	}
	else
		return 0;
}

static int
pop (struct node *head)
{
	struct node *temp;
	int index;
	if (head->next) {
		index = head->next->child_index;
		temp = head->next;
		head->next = temp->next;
		g_free (temp);
	}
	else
		index = -1;
	return index;
}

static struct node*
init_stack ()
{
	struct node *head;
	head = (struct node *) malloc (sizeof (struct node));
	if (head) {
		head->child_index = -1;
		head->next = NULL;
		return head;
	}
	else
		return NULL;
}

static struct node*
trace_path_to_parent (GHashTable *context, char *context_name,
		      GHashTable *component, FILE *log_fp)
{
	struct node *head;
	char *parent_name;
	char *prop = NULL;

	head = init_stack ();
	if (!head) {
		log_msg (LDTP_LOG_CAUSE, "Unable to initialize stack", log_fp);
		return NULL;
	}
	else {
		prop = get_property (component, "child_index", log_fp);
		if (!prop) {
			log_msg (LDTP_LOG_CAUSE, "Unable to get_property", log_fp);
			return NULL;
		}
		push (head, atoi (prop));
		parent_name = get_property (component, "parent", log_fp);
		if (parent_name && context_name)
			g_print ("Parent name: %s - Context name: %s\n", parent_name, context_name);
		while (1) {
			if (!context_name || !parent_name ||
			    !g_utf8_collate (context_name, parent_name) ||
			    g_pattern_match_simple (context_name, parent_name))
				break;
			component = (GHashTable *) get_property (context, parent_name, log_fp);
			if (component) {
				prop = get_property (component, "child_index", log_fp);
				if (!prop) {
					log_msg (LDTP_LOG_CAUSE, "Unable to get_property", log_fp);
					return NULL;
				}
				push (head, atoi (prop));
				parent_name = get_property (component, "parent", log_fp);
				if (parent_name && context_name)
					g_print ("Parent name: %s - Context name: %s\n", parent_name, context_name);
			}
			else
				break;
		}
	}
	return head;
}

int
object_state_contains (Accessible *object, int control_type, FILE *log_fp)
{
	int i;
	int count;
	char msg[256];
	char *name = NULL;
	SPIBoolean object_state;
	AccessibleStateSet *state;

	name = Accessible_getRoleName (object);
	state = Accessible_getStateSet (object);
	count = state_list[control_type].states_count;

	g_print ("State check count: %d - Control type: %d\n", count, control_type);

	for (i = 0; i < count; i++) {
		object_state = AccessibleStateSet_contains (state,
							    state_list[control_type].states [i]);
		if (object_state != 1) {
			g_sprintf (msg, "%s required state %d is not enabled", name, i);
			log_msg (LDTP_LOG_CAUSE, msg, log_fp);
			SPI_freeString (name);
			return -1;
		}
	}
	SPI_freeString (name);
	return 0;
}

/*
  Get accessible handle of the given application
*/
static Accessible*
get_accessible_app_handle (char *app_name)
{
	long i, num_apps;
	Accessible *desktop;

	/*
	  Get SPI handle of desktop 0
	  FIXME: Need to change the hardcoded value of desktop
	*/
	desktop = SPI_getDesktop (0);

	/*
	  Get the total count of childs in the desktop
	*/
	num_apps = Accessible_getChildCount (desktop);
	/*
	  Traverse the list of application handle and if application
	  name matches, then return the handle
	*/
	for (i = 0; i < num_apps; i++) {
		char *name, *descr;
		Accessible *child;

		child = Accessible_getChildAtIndex (desktop, i);
		if (!child)
			continue;
		/*
		  Application name of child handle
		*/
		name = Accessible_getName (child);
		/*
		  Incase of OO.o due to spaces, string compare gives different result.
		  Found because of extra spaces. Now stripped it ;)
		*/
		if (strchr (name, ' '))
			g_strstrip (name);

		/*
		  Description of child handle
		*/
		descr = Accessible_getDescription (child);

		if (name == NULL || app_name == NULL) {
			if (name)
				SPI_freeString (name);
			if (descr)
				SPI_freeString (descr);
			Accessible_unref (child);
			Accessible_unref (desktop);
			return 0;
		}
		g_print ("Accessible application name: %s\n", name);
		if (g_ascii_strcasecmp (name, app_name) == 0) {
			/*
			  Application name matched
			*/
			char *role_name;

			role_name = Accessible_getRoleName (child);
			if (role_name)
				g_print ("Application: %s \t- Role: %s\n", name, role_name);
			SPI_freeString (name);
			SPI_freeString (descr);
			SPI_freeString (role_name);

			Accessible_unref (desktop);
			return child;
		}
		SPI_freeString (descr);
		SPI_freeString (name);
		Accessible_unref (child);
	}
	Accessible_unref (desktop);
	return NULL;
}

/*
 * To get accessible handle of the context (window)
 */
static Accessible*
get_accessible_context_handle (Accessible *app_handle, char *context)
{
	long i;
	long child_count;
	char *child_name  = NULL;
	char *window_name = NULL;
#ifdef ENABLE_LOCALIZATION
	char *temp = NULL;
#endif
	Accessible *child = NULL;

	if (!app_handle)
		return NULL;

	child_count = Accessible_getChildCount (app_handle);
	window_name = Accessible_getName (app_handle);
	if (window_name)
		g_print ("Child count: %ld of %s\n", child_count, window_name);
	SPI_freeString (window_name);
	if (child_count > 0) {
		for (i = 0; i < child_count; i++) {
			child = Accessible_getChildAtIndex (app_handle, i);
			if (!child)
				continue;
			child_name = Accessible_getName (child);
			window_name = get_window_text_in_appmap_format (child_name, Accessible_getRole (child));
			if (last_new_context && last_existing_context) {
				g_print ("%s - %s - %s - %s\n", last_new_context, last_existing_context, context, child_name);
				if (context && child_name && window_name &&
				    (g_utf8_collate (last_existing_context, context) == 0 ||
				     g_pattern_match_simple (last_existing_context, context) ||
				     g_utf8_collate (context, window_name) == 0 ||
				     g_pattern_match_simple (context, window_name))) {
#ifdef ENABLE_LOCALIZATION
					if ((temp = ldtp_compare_with_locale (last_new_context, child_name))) {
						g_free (temp);
						g_free (window_name);
						SPI_freeString (child_name);
						return child;
					}
#else
					if (g_utf8_collate (last_new_context, child_name) == 0 ||
					    g_pattern_match_simple (last_new_context, child_name)) {		     
						g_free (window_name);
						SPI_freeString (child_name);
						return child;
					}
#endif
				}
			}
#ifdef ENABLE_LOCALIZATION
			if (context && child_name && window_name &&
			    (temp = ldtp_compare_with_locale (context, child_name))) {
				g_free (temp);
				g_free (window_name);
				g_print ("Context: %s In_Locale: %s Child_name: %s\n", context, temp, child_name);
				SPI_freeString (child_name);
				return child;
			}
#else
			g_print ("Context: %s Child_name: %s\n", context, child_name);
			if (context && child_name && window_name &&
			    (g_utf8_collate (context, child_name) == 0 ||
			    g_pattern_match_simple (context, child_name) ||
			     g_utf8_collate (context, window_name) == 0 ||
			     g_pattern_match_simple (context, window_name))) {
				g_free (window_name);
				g_print ("Matched - Context: %s Child_name: %s\n", context, child_name);
				SPI_freeString (child_name);
				return child;
			}
#endif
			else {
				g_free (window_name);
				window_name = NULL;
				SPI_freeString (child_name);
				Accessible_unref (child);
			}
		}
	}
	return NULL;
}

/*
  Get accessible handle of the given object
*/
static Accessible*
get_accessible_object_handle (struct node *head, Accessible *context, FILE *log_fp)
{
	long i;
	Accessible *child, *parent;

	if (head) {
		i = pop (head);
		parent = context;
		while (i != -1) {
			child = Accessible_getChildAtIndex (parent, i);
			if (!child) {
				char msg[256];
				char *child_name = NULL;
				char *parent_name = NULL;
				child_name = Accessible_getName (context);
				parent_name = Accessible_getName (parent);
				if (parent_name && child_name)
					g_print ("Parent: %s - Child: %s - %ld\n", parent_name, child_name, i);
				if (parent != context)
					Accessible_unref (parent);
				g_sprintf (msg, "Unable to find child index as available in appmap");
				g_print ("%s\n", msg);
				log_msg (LDTP_LOG_CAUSE, msg, log_fp);
				SPI_freeString (child_name);
				SPI_freeString (parent_name);
				return NULL;
			}
			if (parent != context)
				Accessible_unref (parent);
			parent = child;
			i = pop (head);
		}
		g_free (head);
		return parent;
	}
	else {
		g_print ("Head pointer is NULL\n");
		return NULL;
	}
}

char*
get_relation_name (Accessible *object, long *len)
{
	long i = 0, j = 0;
	int max_accessible = 0;
	char *label_by = NULL;
	AccessibleRelation **relation = NULL;

	relation = Accessible_getRelationSet (object);
	if (!relation) {
		return NULL;
	}
	max_accessible = AccessibleRelation_getNTargets (*relation);
	if (max_accessible <= 0) {
		Accessible_unref (*relation);
		return NULL;
	}
	for (i = 0; relation[i] != NULL; i++) {
		for (j = 0; j < AccessibleRelation_getNTargets (relation[i]); j++) {
			AccessibleRelationType relation_type;
			relation_type = AccessibleRelation_getRelationType (relation[i]);
			if (relation_type == SPI_RELATION_LABELED_BY ||
			    relation_type == SPI_RELATION_CONTROLLED_BY ||
			    relation_type == SPI_RELATION_LABEL_FOR) {
				int k = 0;
				char *name = NULL;
				Accessible *tmp_obj;
				tmp_obj = AccessibleRelation_getTarget (relation[i], i);
				name = Accessible_getName (tmp_obj);
				if (name == NULL) {
					Accessible_unref (tmp_obj);
					for (k = j; relation[k] != NULL; k++)
						Accessible_unref (relation[k]);
					//Accessible_unref (*relation);
					return NULL;
				}
				if (g_ascii_strcasecmp (name, "") == 0) {
					SPI_freeString (name);
					Accessible_unref (tmp_obj);
					for (k = j; relation[k] != NULL; k++)
						Accessible_unref (relation[k]);
					//Accessible_unref (*relation);
					return NULL;
				}
				g_print ("Relation name: %s\n", name);
				label_by = g_strdup (name);
				SPI_freeString (name);
				if (Accessible_isText (tmp_obj) && len != NULL) {
					AccessibleText *text = NULL;
					text = Accessible_getText (tmp_obj);
					*len = AccessibleText_getCharacterCount (text);
					Accessible_unref (text);
				}
				Accessible_unref (tmp_obj);
				for (k = j; relation[k] != NULL; k++)
					Accessible_unref (relation[k]);
				//Accessible_unref (*relation);
				return label_by;
			}
		} // for j
		Accessible_unref (relation[i]);
	} // for i
	//Accessible_unref (*relation);
	return NULL;
} // get_relation_name

static gboolean
is_object_matching (Accessible *object, GHashTable *comp_attributes, FILE *log_fp)
{
	char *accessible_label;
	char *hash_label;
#ifdef ENABLE_LOCALIZATION
	gchar *value;
#else
	gint value;
#endif
	/*
	 * NOTE: Checking if the obtained object is the required one 
	 * with respect to label/label_by only
	 */
	//hash_label = g_hash_table_lookup (comp_attributes, "label");
	hash_label = get_property (comp_attributes, "label", log_fp);
	if (hash_label) {
		accessible_label = Accessible_getName (object);
		g_print ("Label: %s - %s\n", hash_label, accessible_label);
#ifdef ENABLE_LOCALIZATION
		value = ldtp_compare_with_locale (hash_label, accessible_label);
		if (value)
#else
			if (g_utf8_strchr (hash_label, -1, '_') != NULL) {
				gchar *str = NULL;
				str = escape_character (hash_label, '_');
				g_print ("Before: %s - After: %s\n", hash_label, str);
				value = g_utf8_collate (str, accessible_label);
				g_free (str);
			}
			else
				value = g_utf8_collate (hash_label, accessible_label);
		if (value == 0)
#endif
			{
				SPI_freeString (accessible_label);
				g_print ("Object matches\n");
#ifdef ENABLE_LOCALIZATION
				g_free (value);
#endif
				return TRUE;
			}
		else {
			SPI_freeString (accessible_label);
			return FALSE;
		}
	}
	else {
		//hash_label = g_hash_table_lookup (comp_attributes, "label_by");
		hash_label = get_property (comp_attributes, "label_by", log_fp);
		if (hash_label) {
			long len = -1;
			accessible_label = get_relation_name (object, &len);
#ifdef ENABLE_LOCALIZATION
			value = ldtp_compare_with_locale (hash_label, accessible_label);
			if (value)
#else
				if (g_utf8_strchr (hash_label, -1, '_') != NULL) {
					gchar *str = NULL;
					str = escape_character (hash_label, '_');
					g_print ("Before: %s - After: %s\n", hash_label, str);
					value = g_utf8_collate (str, accessible_label);
					g_free (str);
				}
				else
					value = g_utf8_collate (hash_label, accessible_label);
			if (value == 0)
#endif
				{
					g_free (accessible_label);
					g_print ("Object matches\n");
#ifdef ENABLE_LOCALIZATION
					g_free (value);
#endif
					return TRUE;
				}
			else {
				g_free (accessible_label);
				return FALSE;
			}
		}
	}
	return TRUE;
}

static Accessible*
get_child_window_handle (Accessible *app_handle, char *context)
{
	long i;
	char *name = NULL;
	char *window_name = NULL;
	long child_count = 0;
	Accessible *child = NULL;
	GPatternSpec *pattern = NULL;

	/*
	  FIXME: If app_handle is NULL, then get the desktop handle and from there
	  search for window handle instead of directly returning from here
	*/
	if (!app_handle)
		return NULL;
	child_count = Accessible_getChildCount (app_handle);
	pattern = g_pattern_spec_new (context);
	if (!pattern)
		return NULL;

	for (i = 0; i < child_count; i++) {
		child = Accessible_getChildAtIndex (app_handle, i);
		if (!child)
			continue;
		/*
		  Application name of child handle
		*/
		name = Accessible_getName (child);
		if (name)
			window_name = get_window_text_in_appmap_format (name, Accessible_getRole (child));
		if (context && name && window_name)
			g_print ("Search window name: %s - %s - %s\n", context, name, window_name);
		if (name && window_name && context &&
		    (g_utf8_collate (context, name) == 0 ||
		     g_pattern_match_string (pattern, name) ||
		     g_utf8_collate (context, window_name) == 0 ||
		     g_pattern_match_string (pattern, window_name))) {
			g_print ("Window name matched\n");
			g_free (window_name);
			SPI_freeString (name);
			g_pattern_spec_free (pattern);
			return child;
		}
		if (window_name) {
			g_free (window_name);
			window_name = NULL;
		}
		SPI_freeString (name);
		Accessible_unref (child);
	}
	g_pattern_spec_free (pattern);
	return NULL;
}

static Accessible*
get_window_handle (Accessible *app_handle, char *context)
{
	char *name = NULL;
	long i, num_apps;
	Accessible *child = NULL;
	Accessible *desktop = NULL;
	Accessible *window_handle = NULL;
	if (app_handle) {
		window_handle = get_child_window_handle (app_handle, context);
		if (window_handle)
			return window_handle;
	}

	/*
	  Get SPI handle of desktop 0
	  FIXME: Need to change the hardcoded value of desktop
	*/
	desktop = SPI_getDesktop (0);

	/*
	  Get the total count of childs in the desktop
	*/
	num_apps = Accessible_getChildCount (desktop);
	/*
	  Traverse the list of application handle and if application
	  name matches, then return the handle
	*/
	for (i = 0; i < num_apps; i++) {
		child = Accessible_getChildAtIndex (desktop, i);
		if (!child)
			continue;
		/*
		  Application name of child handle
		*/
		name = Accessible_getName (child);
		if (!name)
			continue;
		/*
		  Incase of OO.o due to spaces, string compare gives different result.
		  Found because of extra spaces. Now stripped it ;)
		*/
		if (strchr (name, ' '))
			g_strstrip (name);

		g_print ("Accessible application name: %s\n", name);
		window_handle = get_child_window_handle (child, context);
		if (window_handle) {
			SPI_freeString (name);
			Accessible_unref (child);
			Accessible_unref (desktop);
			return window_handle;
		}
		SPI_freeString (name);
		Accessible_unref (child);
	}
	Accessible_unref (desktop);
	return NULL;
}

void
update_cur_window_appmap_handle (LDTPClientContext* cctxt,
				 LDTPError** err)
{
	char *context = NULL;
	char *window_name = NULL;
	char *parent_name = NULL;
	Accessible *child = NULL;
	Accessible *parent = NULL;
	GHashTable *new_hashtable = NULL;

	context = (char *) cctxt->req->context;
	if (!context) {
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		g_print ("%s", (*err)->message);
		return;
	}
	child = get_window_handle (cctxt->app_handle, context);
	if (!child) {
		g_print ("Window %s not open\n", context);
		*err = ldtp_error_get_error (LDTP_ERROR_WIN_NOT_OPEN);
		return;
	}
	/*
	  Unable to find using the new context based on the dynamic context generated
	*/
	parent = Accessible_getParent (child);
	if (parent) {
		AccessibleRole role;
		parent_name = Accessible_getName (parent);
		if (parent_name)
			g_print ("Parent name: %s\n", parent_name);
		role = Accessible_getRole (parent) == SPI_ROLE_APPLICATION ? SPI_ROLE_APPLICATION : Accessible_getRole (child);
		window_name = get_window_text_in_appmap_format (parent_name, role);
		if (window_name && parent_name)
			g_print ("Window name - Stripped: %s - %s\n", window_name, parent_name);
		SPI_freeString (parent_name);
		Accessible_unref (parent);
	}
	else
		g_print ("Unable to get parent\n");
	pthread_mutex_lock (&mutex);
	new_hashtable = do_remap (child, window_name, NULL, cctxt->locale_set);
	pthread_mutex_unlock (&mutex);
	if (window_name) {
		g_free (window_name);
		window_name = NULL;
	}
	if (new_hashtable) {
		char *name = NULL;
		if (!cctxt->app_map) {
			g_print ("Appmap hashtable not initalized\n");
			cctxt->app_map = g_hash_table_new (&g_str_hash, &g_str_equal);
		}
		name = Accessible_getName (child);
		if (name) {
			window_name = get_window_text_in_appmap_format (name, Accessible_getRole (child));
			if (name && window_name)
				g_print ("Window name: %s - Stripped: %s\n", window_name, name);
			SPI_freeString (name);
		}
		Accessible_unref (child);
		if (window_name) {
			pthread_mutex_lock (&mutex);
			g_hash_table_insert (cctxt->app_map, g_strdup (window_name), new_hashtable);
			pthread_mutex_unlock (&mutex);
			g_free (window_name);
			window_name = NULL;
		}
		else {
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_UPDATE_APPMAP);
			g_print ("DEBUG: Window name NULL\n");
			return;
		}
	}
	else {
		Accessible_unref (child);
		g_print ("DEBUG: New hash table NULL\n");
		*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_UPDATE_APPMAP);
		return;
	}
	/*
	  Enable it for debugging purpose
	*/
	g_hash_table_foreach (cctxt->app_map, (GHFunc)&print_context, NULL);
	*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
}

/*
  Get gui handle based on context and compoment
*/
LDTPGuiHandle*
ldtp_gui_get_gui_handle (LDTPClientContext* cctxt,
			 LDTPError** err)
{
	char msg[256]     = "";
	gchar *context    = NULL;
	char *app_name    = NULL;
	char *component   = NULL;
	char *window_name = NULL;
	struct node *head;
	AccessibleRole class;
	Accessible *object_handle  = NULL;
	Accessible *context_handle = NULL;
	LDTPGuiHandle* ctxt_handle = NULL;

	GHashTable *cur_window;
	GHashTable *cur_component;

	if (!cctxt || !cctxt->req->context) {
		/* FIXME: Error handling */
		*err = ldtp_error_get_error (LDTP_ERROR_ARGUMENT_NULL);
		log_msg (LDTP_LOG_CAUSE, (*err)->message, cctxt->log_fp);
		return NULL;
	}

	context = (char *) cctxt->req->context;
	component = (char *) cctxt->req->component;
	/*
	  For guiexist, waittillguiexist functions, both context and component are same
	*/
	if (!cctxt->req->component)
		component = context;

	if (!cctxt->app_map) {
		update_cur_window_appmap_handle (cctxt, err);
		if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
			g_sprintf (msg, "Unable to update context: %s in appmap", context);
			goto error;
		}
	}

	cur_window = get_object_def (cctxt->app_map, context, cctxt->log_fp);
	/*
	  If window info, not available in appmap hash table, try to remap it
	  iff, the window exist
	*/
	if (!cur_window) {
		update_cur_window_appmap_handle (cctxt, err);
		if (*err && (*err)->err_code != LDTP_ERROR_SUCCESS) {
			g_sprintf (msg, "Unable to update context: %s in appmap", context);
			goto error;
		}
		cur_window = get_object_def (cctxt->app_map, context, cctxt->log_fp);
		if (!cur_window) {
			char *name;
			Accessible *child;
			/*
			  If context is given without window type, then use dynamically created window name
			  Example: if context is Replace, instead of dlgReplace (in gedit app, replace dialog)
			  then we will try to search using our dynamic context dlgReplace
			*/
			child = get_window_handle (cctxt->app_handle, context);
			if (!child) {
				g_print ("Window %s not open\n", context);
				g_sprintf (msg, "Window %s not open", context);
				*err = ldtp_error_get_error (LDTP_ERROR_WIN_NOT_OPEN);
				goto error;
			}
			name = Accessible_getName (child);
			window_name = get_window_text_in_appmap_format (name, Accessible_getRole (child));
			Accessible_unref (child);
			if (window_name) {
				gboolean tmpflag = FALSE;
				if (cctxt->req->context == component)
					tmpflag = TRUE;
				if (cctxt->req->context)
					g_free (cctxt->req->context);
				cctxt->req->context = g_strdup (window_name);
				context = (char *) cctxt->req->context;
				/* 
				   In case of waittillguiexist and waittillguinotexist,
				   both context and component members of the ClientContext 
				   structure points to same string and they don't need 
				   class_id.
				*/
				if (tmpflag)
					component = context;
				if (context)
					cur_window = get_object_def (cctxt->app_map, context, cctxt->log_fp);
				g_free (window_name);
				window_name = NULL;
			}
		}
		if (!cur_window) {
			g_sprintf (msg, "Unable to find window name: %s in appmap", context);
			*err = ldtp_error_get_error (LDTP_ERROR_WIN_NAME_NOT_FOUND_IN_APPMAP);
			goto error;
		}
	}
	cur_component = get_object_def (cur_window, context, cctxt->log_fp);
	if (!cur_component) {
		g_sprintf (msg, "%s - %d - Unable to find object name: %s in appmap", __FILE__, __LINE__, context);
		*err = ldtp_error_get_error (LDTP_ERROR_OBJ_NAME_NOT_FOUND_IN_APPMAP);
		goto error;
	}
	app_name = get_property (cur_component, "parent", cctxt->log_fp);
	g_print ("app_name: %s\n", app_name);
	window_name = get_property (cur_component, "label", cctxt->log_fp);
	g_print ("window_name: %s\n", window_name);

	cur_component = get_object_def (cur_window, component, cctxt->log_fp);

	if (!cur_component) {
		update_cur_window_appmap_handle (cctxt, err);
		cur_window = get_object_def (cctxt->app_map, context, cctxt->log_fp);
		cur_component = get_object_def (cur_window, component, cctxt->log_fp);

		if (!cur_component) {
			g_sprintf (msg, "%s - %d - Unable to find object name: %s in appmap", __FILE__, __LINE__, component);
			*err = ldtp_error_get_error (LDTP_ERROR_OBJ_NAME_NOT_FOUND_IN_APPMAP);
			goto error;
		}
	}
  
	/*
	  Update the ClientContext structure with application 
	  handle, iff, it wasn't already updated.
	  FIXME: We should think about scenarios that involve
	  multiple-app-maps and multiple-application-handles.
	*/
	//if (!cctxt->app_handle) {
	if (app_name)
		g_print ("Appname: %s\n", app_name);
	if (cctxt->app_handle)
		Accessible_unref (cctxt->app_handle);
	cctxt->app_handle = get_accessible_app_handle (app_name);
	if (!cctxt->app_handle) {
		g_sprintf (msg, "Application: %s not running", app_name);
		*err = ldtp_error_get_error (LDTP_ERROR_APP_NOT_RUNNING);
		goto error;
	}
		//}
	context_handle = get_accessible_context_handle (cctxt->app_handle, window_name);
	if (!context_handle)
		context_handle = get_accessible_context_handle (cctxt->app_handle, context);
	if (!context_handle) {
		if (g_utf8_collate (context, component)) {
			g_sprintf (msg, "Window: %s not opened", context);
			*err = ldtp_error_get_error (LDTP_ERROR_WIN_NOT_OPEN);
		}
		else
			*err = ldtp_error_get_error (LDTP_ERROR_UNABLE_TO_GET_CONTEXT_HANDLE);
		goto error;
	}

	ctxt_handle = calloc (1, sizeof (LDTPGuiHandle));
	ctxt_handle->handle = context_handle;

	g_print ("Window: %s - Object: %s\n", context, component);

	/* 
	   In case of waittillguiexist and waittillguinotexist,
	   both context and component members of the ClientContext 
	   structure points to same string and they don't need 
	   class_id.
	*/
	if (context && component &&
	    !g_utf8_collate ((const gchar *)context, (const gchar *)component))
		return ctxt_handle;

	head = trace_path_to_parent (cur_window, context, cur_component, cctxt->log_fp);
	object_handle = get_accessible_object_handle (head, context_handle, cctxt->log_fp);
	Accessible_unref (context_handle);
	if (!object_handle) {
		/*
		  To avoid runtime remap function to be called, let us call remap here
		*/
		update_cur_window_appmap_handle (cctxt, err);
		context_handle = get_accessible_context_handle (cctxt->app_handle, window_name);
		if (!context_handle)
			context_handle = get_accessible_context_handle (cctxt->app_handle, context);
		if (context_handle) {
			cur_window = get_object_def (cctxt->app_map, context, cctxt->log_fp);
			cur_component = get_object_def (cur_window, component, cctxt->log_fp);
			head = trace_path_to_parent (cur_window, context, cur_component, cctxt->log_fp);
			object_handle = get_accessible_object_handle (head, context_handle, cctxt->log_fp);
			Accessible_unref (context_handle);
		}
	}
	if (!object_handle) {
		if (component)
			g_sprintf (msg, "Unable to get object handle %s", component);
		*err = ldtp_error_get_error (LDTP_ERROR_GET_OBJ_HANDLE_FAILED);
		goto error;
	}

	if (!is_object_matching (object_handle, cur_component, cctxt->log_fp)) {
		char *name = NULL;
		name = Accessible_getName (object_handle);
		if (name) {
			g_sprintf (msg, "Object information: %s does not match with appmap entry", name);
			SPI_freeString (name);
		}
		Accessible_unref (object_handle);
		*err = ldtp_error_get_error (LDTP_ERROR_OBJ_INFO_MISMATCH);
		goto error;
	}
	ctxt_handle->class_id = 0;
	class = Accessible_getRole (object_handle);

	if (class == SPI_ROLE_EXTENDED) {
		char *name;
		name = Accessible_getRoleName (object_handle);
		if (g_ascii_strcasecmp (name, "calendar view") == 0)
			ctxt_handle->class_id = CALENDAR_VIEW;
		else if (g_ascii_strcasecmp (name, "calendar event") == 0)
			ctxt_handle->class_id = CALENDAR_EVENT;
		SPI_freeString (name);
	}
	else
		ctxt_handle->class_id = class;
  
	ctxt_handle->handle = object_handle;
	*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
	return ctxt_handle;

 error:
	if (g_ascii_strcasecmp (msg, "") != 0) {
		g_print ("%s\n", msg);
		log_msg (LDTP_LOG_ERROR, msg, cctxt->log_fp);
	}
	return NULL;
}

static Accessible*
get_object_handle (Accessible *accessible, AccessibleRole role)
{
	long i;
	long count;
	Accessible *child, *parent;
	
	count = Accessible_getChildCount (accessible);
	for (i = 0; i < count; i++) {
		child = Accessible_getChildAtIndex (accessible, i);

		if (!child)
			continue;	

		parent = child;
		if (Accessible_getRole (child) == role)
			return child;
		else {
			Accessible *parent = NULL;
			parent = child;
			child = get_object_handle (child, role);
			/*
			  Unref local parent
			*/
			if (parent)
				Accessible_unref (parent);
			if (child != NULL)
				return child;
		}
		Accessible_unref (parent);
	}
	return NULL;
}

Accessible *
get_list_handle (Accessible *accessible)
{
	return get_object_handle (accessible, SPI_ROLE_LIST);
}

Accessible *
get_text_handle (Accessible *accessible)
{
	return get_object_handle (accessible, SPI_ROLE_TEXT);
}

Accessible *
get_menu_handle (Accessible *accessible)
{
	return get_object_handle (accessible, SPI_ROLE_MENU);
}

int 
get_object_type (Accessible *accessible)
{
	int object_type = 0;
	AccessibleRole role;
	role = Accessible_getRole (accessible);

	if (role == SPI_ROLE_LIST)
		object_type = SPI_ROLE_LIST;
	else if (role == SPI_ROLE_MENU)
		object_type = SPI_ROLE_MENU;
	else if (role == SPI_ROLE_TEXT)
		object_type = SPI_ROLE_TEXT;
	else if (role == SPI_ROLE_TOGGLE_BUTTON)
		object_type = SPI_ROLE_TOGGLE_BUTTON;
	else if (role == SPI_ROLE_TABLE_CELL)
		object_type = SPI_ROLE_TABLE_CELL;
	else if (role == SPI_ROLE_EXTENDED) {
		char *role_name;
		role_name = Accessible_getRoleName (accessible);
		if (g_ascii_strcasecmp (role_name, "calendar event") == 0)
			object_type = CALENDAR_EVENT;
		SPI_freeString (role_name);
	}
	return object_type;
}

int 
get_child_object_type (Accessible *accessible)
{
	int i;
	int num_child;
	int child_type;
	AccessibleRole role;
	Accessible *child;
                                                                                
	num_child = Accessible_getChildCount (accessible);
	for (i = 0; i < num_child; i++) {
		child = Accessible_getChildAtIndex (accessible, i);
		if (!child)
			continue;
                                                                                
		role = Accessible_getRole (child);
		if (role == SPI_ROLE_MENU) {
			Accessible_unref (child);
			return SPI_ROLE_MENU;
		}
		else if (role == SPI_ROLE_LIST) {
			Accessible_unref (child);
			return SPI_ROLE_LIST;
		}
                                                                                
		child_type = get_child_object_type (child);
		if (child_type != 0) {
			Accessible_unref (child);
			return child_type;
		}
		Accessible_unref (child);
	}
	return 0;
}

int 
set_new_context (char *existing_context, char *new_context)
{
	if (last_new_context) {
		g_free (last_new_context);
		last_new_context = NULL;
	}
	if (last_existing_context) {
		g_free (last_existing_context);
		last_existing_context = NULL;
	}
	last_new_context = g_strdup (new_context);
	last_existing_context = g_strdup (existing_context);
	return 1;
}

int 
release_last_context ()
{
	if (last_new_context) {
		g_free (last_new_context);
		last_new_context = NULL;
	}
	if (last_existing_context) {
		g_free (last_existing_context);
		last_existing_context = NULL;
	}
	return 1;
}

void
ldtp_gui_free_gui_handle (LDTPGuiHandle* ui_handle)
{
	if (!ui_handle)
		return;
	Accessible_unref (ui_handle->handle);
	g_free (ui_handle);
}

void
ldtp_gui_wait_till_gui_exist (LDTPClientContext* cctxt, LDTPError** err)
{
	int time_wait = 30; // Max wait for 30 seconds
	char *env_sleep = NULL;
	time_t start, cur;
	LDTPGuiHandle *accessible = NULL;

	/*
	  Copied from LTFX source code
	*/
	start = cur = time (NULL);

	if ((env_sleep = getenv ("GUI_TIMEOUT")) != NULL) {
		time_wait = atoi (env_sleep);
		if (time_wait == 0)
			time_wait = 30;
	}

	while ((difftime (cur, start) < time_wait) && (accessible == NULL)) {
		ldtp_nsleep (0, 15000);

		cur = time (NULL);
		accessible = ldtp_gui_get_gui_handle (cctxt, err);
	}

	if (accessible == NULL)
		*err = ldtp_error_get_error (LDTP_ERROR_GUI_NOT_EXIST);
	else {
		Accessible_unref (accessible->handle);
		*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
	}
}

void
ldtp_gui_wait_till_gui_not_exist (LDTPClientContext* cctxt, LDTPError** err)
{
	int time_wait = 30; // Max wait for 30 seconds
	char *env_sleep = NULL;
	time_t start, cur;
	LDTPGuiHandle *accessible = NULL;

	/*
	  Copied from LTFX source code
	*/
	start = cur = time (NULL);

	if ((env_sleep = getenv ("GUI_TIMEOUT")) != NULL) {
		time_wait = atoi (env_sleep);
		if (time_wait == 0)
			time_wait = 30;
	}

	accessible = ldtp_gui_get_gui_handle (cctxt, err);

	while ((difftime (cur, start) < time_wait) && (accessible != NULL)) {
		ldtp_nsleep (0, 15000);

		cur = time (NULL);
		if (accessible) {
			Accessible_unref (accessible->handle);
			accessible = NULL;
		}
		accessible = ldtp_gui_get_gui_handle (cctxt, err);
	}

	if (accessible == NULL)
		*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
	else {
		Accessible_unref (accessible->handle);
		*err = ldtp_error_get_error (LDTP_ERROR_GUI_EXIST);
	}
}

void
ldtp_gui_gui_exist (LDTPClientContext* cctxt, LDTPError** err)
{
	LDTPGuiHandle *accessible = NULL;

	accessible = ldtp_gui_get_gui_handle (cctxt, err);
	if (accessible == NULL)
		*err = ldtp_error_get_error (LDTP_ERROR_GUI_NOT_EXIST);
	else {
		Accessible_unref (accessible->handle);
		*err = ldtp_error_get_error (LDTP_ERROR_SUCCESS);
	}
}
