/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Veerapuram Varadhan <v.varadhan@gmail.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <glib.h>

#include "ldtp-server.h"
#include "ldtp-logger.h"

static struct sockaddr_un myaddr;   // server address
int script_listener = 0;                  // listening socket descriptor
int record_listener = 0;                  // listening socket descriptor

char*
get_tmp_file (int server_type)
{
	if (server_type == LDTP_SCRIPT_SERVER)
		return g_strdup_printf ("/tmp/ldtp-%s-%s", getenv ("USER"), getenv ("DISPLAY"));
	else if (server_type == LDTP_RECORD_SERVER)
		return g_strdup_printf ("/tmp/ldtp-record-%s-%s", getenv ("USER"), getenv ("DISPLAY"));
	else
		return NULL;
}

int
init_ldtp_server (int server_type)
{
	int yes = 1;        // for setsockopt() SO_REUSEADDR, below
	gchar* tmpfile = NULL;
	int listener = 0;

	// get the listener
	if ((listener = socket (AF_UNIX, SOCK_STREAM, 0)) == -1) {
		ldtp_log ("ERROR:socket() failed with \"%s\"\n", strerror(errno));
		exit (-1);
	}

	if (server_type == LDTP_SCRIPT_SERVER)
		script_listener = listener;
	else if (server_type == LDTP_SCRIPT_SERVER)
		record_listener = listener;

	// bind
	tmpfile = get_tmp_file (server_type);
	/*
	  If file already exist, we need to unlink it, otherwise we will get bind error
	*/
	unlink (tmpfile);
	strcpy (myaddr.sun_path, tmpfile);
	g_free (tmpfile);
	myaddr.sun_family = AF_UNIX;
	if (setsockopt (listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int)) == -1) {
		ldtp_log ("ERROR:setsockopt() failed with \"%s\"\n", strerror (errno));
		exit (-1);
	}

	if (bind (listener, (struct sockaddr *) &myaddr, sizeof (myaddr)) == -1) {
		ldtp_log ("ERROR:bind() failed with \"%s\"\n", strerror (errno));
		exit (-1);
	}

	// listen
	if (listen (listener, 10) == -1) {
		ldtp_log ("ERROR:listen() failed with \"%s\"\n", strerror (errno));
		exit (-1);
	}
	return listener;
}

int
get_server_socket (int server_type)
{
	if (server_type == LDTP_SCRIPT_SERVER)
		return script_listener;
	else if (server_type == LDTP_RECORD_SERVER)
		return record_listener;
	else
		return 0;
}
