/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/*
 * Linux Desktop Testing Project http://ldtp.freedesktop.org
 *
 * Author:
 *    Nagappan A <nagappan@gmail.com>
 *
 * Copyright 2004 - 2006 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <glib.h>

#include "ldtp-state.h"

static GHashTable* state_id_table = NULL;
 
static LDTPState*
ldtp_state_new (LDTPStateId id, const char* state_name)
{
	LDTPState *state = calloc (1, sizeof (LDTPState));
	state->state_id = id;
	state->state_name = g_strdup (state_name);
	return state;
}

static void
ldtp_state_free (LDTPState *state)
{
	if (!state)
		return;

	g_free (state->state_name);
	g_free (state);
	state = NULL;
}

static void
ldtp_state_init_table ()
{
	int i = 0;
	LDTPState *state[] = 
		{
			ldtp_state_new ( LDTP_STATE_ACTIVE, "active" ),
			ldtp_state_new ( LDTP_STATE_ARMED, "armed" ),
			ldtp_state_new ( LDTP_STATE_BUSY, "busy" ),
			ldtp_state_new ( LDTP_STATE_CHECKED, "checked" ),
			ldtp_state_new ( LDTP_STATE_COLLAPSED, "collapsed" ),
			ldtp_state_new ( LDTP_STATE_DEFUNCT, "defunct" ),
			ldtp_state_new ( LDTP_STATE_EDITABLE, "editable" ),
			ldtp_state_new ( LDTP_STATE_ENABLED, "enabled" ),
			ldtp_state_new ( LDTP_STATE_EXPANDABLE, "expandable" ),
			ldtp_state_new ( LDTP_STATE_EXPANDED, "expanded" ),
			ldtp_state_new ( LDTP_STATE_FOCUSABLE, "focusable" ),
			ldtp_state_new ( LDTP_STATE_FOCUSED, "focused" ),
			ldtp_state_new ( LDTP_STATE_HORIZONTAL, "horizontal" ),
			ldtp_state_new ( LDTP_STATE_ICONIFIED, "confined" ),
			ldtp_state_new ( LDTP_STATE_MODAL, "modal" ),
			ldtp_state_new ( LDTP_STATE_MULTILINE, "multiline" ),
			ldtp_state_new ( LDTP_STATE_MULTISELECTABLE, "multiselectable" ),
			ldtp_state_new ( LDTP_STATE_OPAQUE, "opaque" ),
			ldtp_state_new ( LDTP_STATE_PRESSED, "pressed" ),
			ldtp_state_new ( LDTP_STATE_RESIZABLE, "resizable" ),
			ldtp_state_new ( LDTP_STATE_SELECTABLE, "selectable" ),
			ldtp_state_new ( LDTP_STATE_SELECTED, "selected" ),
			ldtp_state_new ( LDTP_STATE_SENSITIVE, "sensitive" ),
			ldtp_state_new ( LDTP_STATE_SHOWING, "showing" ),
			ldtp_state_new ( LDTP_STATE_SINGLELINE, "singleline" ),
			ldtp_state_new ( LDTP_STATE_STALE, "stale" ),
			ldtp_state_new ( LDTP_STATE_TRANSIENT, "transient" ),
			ldtp_state_new ( LDTP_STATE_VERTICAL, "vertical" ),
			ldtp_state_new ( LDTP_STATE_VISIBLE, "visible" ),
			ldtp_state_new ( LDTP_STATE_MANAGESDESCENDANTS, "mangesdescendants" ),
			ldtp_state_new ( LDTP_STATE_INDETERMINATE, "indenterminate" ),
			/*ldtp_state_new ( LDTP_STATE_TRUNCATED, "truncated" ),
			ldtp_state_new ( LDTP_STATE_REQUIRED, "required" ),
			ldtp_state_new ( LDTP_STATE_INVALIDENTRY, "invalidentry" ),
			ldtp_state_new ( LDTP_STATE_SUPPORTSAUTOCOMPLETION, "supportsautocompletion" ),
			ldtp_state_new ( LDTP_STATE_SELECTABLETEXT, "selectabletext" ),
			ldtp_state_new ( LDTP_STATE_ISDEFAULT, "isdefault" ),
			ldtp_state_new ( LDTP_STATE_VISITED, "visited" ),*/
			NULL
		};

	state_id_table = g_hash_table_new_full (&g_str_hash, &g_str_equal, NULL,  (GDestroyNotify)&ldtp_state_free);
	while (state[i]) {
		g_hash_table_insert (state_id_table, state[i]->state_name, state[i]); 
		i++;
	}
}
void
ldtp_state_destroy_state_table (void)
{
	g_hash_table_destroy (state_id_table);
}

LDTPStateId
ldtp_state_get_state_id (const char* state_name)
{
	LDTPState *state;
	if (!state_name)
		return LDTP_STATE_INVALID;
	if (!state_id_table)
		ldtp_state_init_table ();
	g_print ("Get state id: %s\n", state_name);
	state = g_hash_table_lookup (state_id_table, state_name);
	return state ? state->state_id : LDTP_STATE_INVALID;
}
